#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Tests for the R&S RTO1024 digital storage oscilloscope device class.
"""

import pytest

from hvl_ccb.dev.rs_rto1024 import (
    RTO1024,
    RTO1024Config,
    RTO1024Error,
    RTO1024VisaCommunicationConfig,
)
from masked_comm.visa import MaskedVisaCommunication


@pytest.fixture(scope="module")
def com_config():
    return {
        "interface_type": RTO1024VisaCommunicationConfig.InterfaceType.TCPIP_INSTR,
        "host": "127.0.0.1",
        "open_timeout": 10,
        "timeout": 50,
    }


@pytest.fixture(scope="module")
def dev_config():
    return {
        "waveforms_path": "C:\\Data\\DavidGraber\\02_waveforms",
        "settings_path": "C:\\Data\\DavidGraber\\01_settings",
        "backup_path": "D:\\backups",
        "spoll_interval": 0.01,
        "spoll_start_delay": 0,
        "command_timeout_seconds": 0.05,
        "wait_sec_short_pause": 0.01,
        "wait_sec_post_acquisition_start": 0.01,
    }


@pytest.fixture
def testdev(com_config, dev_config):
    com = MaskedVisaCommunication(com_config)
    dev = RTO1024(com, dev_config)
    dev.SPOLL_INTERVAL = 0.01
    dev.SPOLL_START_DELAY = 0
    assert dev is not None
    dev.start()
    assert dev.com.get_written() == "&GTR"
    assert dev.com.get_written() == "*RST"
    assert dev.com.get_written() == "*CLS"
    assert dev.com.get_written() == "SYST:DISP:UPD ON"
    assert dev.com.get_written() == "*ESE 127"
    yield dev
    dev.stop()


def test_dev_config(dev_config) -> None:
    config = RTO1024Config(**dev_config)
    for key, value in dev_config.items():
        assert getattr(config, key) == value


@pytest.mark.parametrize(
    "wrong_config_dict",
    [
        {"spoll_interval": 0},
        {"spoll_interval": -1},
        {"spoll_start_delay": -1},
        {"command_timeout_seconds": 0},
        {"wait_sec_short_pause": 0},
        {"wait_sec_short_pause": -1},
        {"wait_sec_enable_history": 0},
        {"wait_sec_post_acquisition_start": 0},
    ],
)
def test_invalid_config_dict(dev_config, wrong_config_dict) -> None:
    invalid_config = dict(dev_config)
    invalid_config.update(wrong_config_dict)
    with pytest.raises(ValueError):
        RTO1024Config(**invalid_config)


def test_instantiation(com_config, dev_config) -> None:
    dev = RTO1024(com_config, dev_config)
    assert dev is not None

    copy_config = dict(com_config)
    copy_config.pop("interface_type")
    dev = RTO1024(copy_config, dev_config)
    assert dev is not None


def test_stop(testdev: RTO1024) -> None:
    # case polling: send stopping instructions
    testdev.stop()
    assert testdev.com.get_written() == "*ESE 0"
    assert testdev.com.get_written() == "*SRE 0"
    assert testdev.com.get_written() == "&DCL"
    assert testdev.com.get_written() == "&GTL"
    # case not polling: do nothing, just log a warning
    testdev.stop()


def test_local_display(testdev: RTO1024) -> None:
    testdev.local_display(True)
    assert testdev.com.get_written() == "SYST:DISP:UPD ON"

    testdev.local_display(False)
    assert testdev.com.get_written() == "SYST:DISP:UPD OFF"


def test_set_acquire_length(testdev: RTO1024) -> None:
    testdev.set_acquire_length(10)
    assert testdev.com.get_written() == "TIMebase:RANGe 10"


def test_get_acquire_length(testdev: RTO1024) -> None:
    testdev.com.put_name("TIMebase:RANGe?", "5.00E-6")
    assert testdev.get_acquire_length() == 5e-6


def test_set_reference_point(testdev: RTO1024) -> None:
    testdev.set_reference_point(50)
    assert testdev.com.get_written() == "TIMebase:REFerence 50"


def test_get_reference_point(testdev: RTO1024) -> None:
    testdev.com.put_name("TIMebase:REFerence?", "20")
    assert testdev.get_reference_point() == 20


def test_set_repetitions(testdev: RTO1024) -> None:
    testdev.set_repetitions(10)
    assert testdev.com.get_written() == "ACQuire:COUNt 10"


def test_get_repetitions(testdev: RTO1024) -> None:
    testdev.com.put_name("ACQuire:COUNt?", "100")
    assert testdev.get_repetitions() == 100


def test_set_channel_state(testdev: RTO1024) -> None:
    testdev.set_channel_state(2, True)
    assert testdev.com.get_written() == "CHANnel2:STATe ON"

    testdev.set_channel_state(3, False)
    assert testdev.com.get_written() == "CHANnel3:STATe OFF"


def test_get_channel_state(testdev: RTO1024) -> None:
    testdev.com.put_name("CHANnel1:STATe?", "1")
    assert testdev.get_channel_state(1)


def test_set_channel_scale(testdev: RTO1024) -> None:
    testdev.set_channel_scale(1, 1e-3)
    assert testdev.com.get_written() == "CHANnel1:SCALe 0.001"


def test_get_channel_scale(testdev: RTO1024) -> None:
    testdev.com.put_name("CHANnel1:SCALe?", "0.001")
    assert testdev.get_channel_scale(1) == 0.001


def test_set_channel_range(testdev: RTO1024) -> None:
    testdev.set_channel_range(2, 1.5)
    assert testdev.com.get_written() == "CHANnel2:RANGe 1.500"


def test_get_channel_range(testdev: RTO1024) -> None:
    testdev.com.put_name("CHANnel1:RANGe?", "0.01")
    assert testdev.get_channel_range(1) == 0.01


def test_set_channel_position(testdev: RTO1024) -> None:
    testdev.set_channel_position(4, 4)
    assert testdev.com.get_written() == "CHANnel4:POSition 4.00"


def test_get_channel_position(testdev: RTO1024) -> None:
    testdev.com.put_name("CHANnel1:POSition?", "1.5")
    assert testdev.get_channel_position(1) == 1.5


def test_set_channel_offset(testdev: RTO1024) -> None:
    testdev.set_channel_offset(4, 4)
    assert testdev.com.get_written() == "CHANnel4:OFFSet 4.000"


def test_get_channel_offset(testdev: RTO1024) -> None:
    testdev.com.put_name("CHANnel1:OFFSet?", "0.6")
    assert testdev.get_channel_offset(1) == 0.6


def test_set_trigger_source(testdev: RTO1024) -> None:
    testdev.set_trigger_source(1)  # to channel 1
    assert testdev.com.get_written() == "TRIGger1:SOURce CHAN1"

    testdev.set_trigger_source(3, 2)  # to channel 3, B-Event
    assert testdev.com.get_written() == "TRIGger2:SOURce CHAN3"


def test_set_trigger_level(testdev: RTO1024) -> None:
    testdev.set_trigger_level(1, 2e-3)
    assert testdev.com.get_written() == "TRIGger1:LEVel1 0.002"


def test_set_trigger_mode(testdev: RTO1024) -> None:
    testdev.set_trigger_mode(RTO1024.TriggerModes.AUTO)
    assert testdev.com.get_written() == "TRIGger1:MODE AUTO"

    testdev.set_trigger_mode(RTO1024.TriggerModes.NORMAL)
    assert testdev.com.get_written() == "TRIGger1:MODE NORMAL"

    testdev.set_trigger_mode(RTO1024.TriggerModes.FREERUN)
    assert testdev.com.get_written() == "TRIGger1:MODE FREERUN"

    testdev.set_trigger_mode("Auto")
    assert testdev.com.get_written() == "TRIGger1:MODE AUTO"

    testdev.set_trigger_mode("normal")
    assert testdev.com.get_written() == "TRIGger1:MODE NORMAL"

    testdev.set_trigger_mode("FREERUN")
    assert testdev.com.get_written() == "TRIGger1:MODE FREERUN"

    with pytest.raises(RTO1024Error):
        testdev.set_trigger_mode("myunknownmode")

    assert testdev.com.get_written() is None


def test_file_copy(testdev: RTO1024) -> None:
    testdev.com.stb = 32
    testdev.file_copy("C:\\Data\\test.txt", "D:\\test.txt")
    testdev.com.stb = 0

    assert testdev.com.get_written() == "*CLS"
    assert (
        testdev.com.get_written() == "MMEMory:COPY 'C:\\Data\\test.txt', 'D:\\test.txt'"
    )
    assert testdev.com.get_written() == "*OPC"

    # test timeout
    with pytest.raises(RTO1024Error):
        testdev.file_copy("C:\\Data\\test.txt", "D:\\test.txt")


def test_backup_waveform(testdev: RTO1024) -> None:
    testdev.com.stb = 32
    testdev.backup_waveform("test")
    assert testdev.com.get_written() == "*CLS"
    assert (
        testdev.com.get_written()
        == "MMEMory:COPY 'C:\\Data\\DavidGraber\\02_waveforms\\test.Wfm.bin', "
        "'D:\\backups\\test.Wfm.bin'"
    )
    assert testdev.com.get_written() == "*OPC"
    assert testdev.com.get_written() == "*CLS"
    assert (
        testdev.com.get_written()
        == "MMEMory:COPY 'C:\\Data\\DavidGraber\\02_waveforms\\test.bin', "
        "'D:\\backups\\test.bin'"
    )
    assert testdev.com.get_written() == "*OPC"
    testdev.com.stb = 0


def test_list_directory(testdev: RTO1024) -> None:
    testdev.com.put_name(
        "MMEMory:CATalog? 'C:/Data'",
        '142422792,147771314176,".,DIR,0","..,DIR,0","DavidGraber,DIR,0",'
        '"Myriam,DIR,0","test.Wfm.bin,BIN,142422792"',
    )

    dir_list = testdev.list_directory("C:/Data")
    expected = [
        ["DavidGraber", "DIR", "0"],
        ["Myriam", "DIR", "0"],
        ["test.Wfm.bin", "BIN", "142422792"],
    ]

    assert dir_list == expected


def test_save_waveform_history(testdev: RTO1024) -> None:
    testdev.com.put_name(
        f"MMEMory:CATalog? '{testdev.config.waveforms_path}'",
        '142422792,147771314176,".,DIR,0","..,DIR,0","DavidGraber,DIR,0",'
        '"Myriam,DIR,0","test.Wfm.bin,BIN,142422792","test.bin,BIN,1234"',
    )
    testdev.com.stb = 32
    testdev.save_waveform_history("test", 1)
    testdev.com.stb = 0
    assert testdev.com.get_written() == "EXPort:WAVeform:FASTexport ON"
    assert testdev.com.get_written() == "CHAN:HIST ON"
    assert testdev.com.get_written() == "SYST:DISP:UPD OFF"
    assert testdev.com.get_written() == "EXPort:WAVeform:MULTichannel OFF"
    assert testdev.com.get_written() == "EXPort:WAVeform:SOURce C1W1"
    assert (
        testdev.com.get_written()
        == "EXPort:WAVeform:NAME 'C:\\Data\\DavidGraber\\02_waveforms\\test.bin'"
    )
    assert testdev.com.get_written() == "EXPort:WAVeform:DLOGging ON"
    assert testdev.com.get_written() == "*CLS"
    assert testdev.com.get_written() == "CHANnel:HISTory:PLAY"
    assert testdev.com.get_written() == "*OPC"
    assert testdev.com.get_written() == "EXPort:WAVeform:FASTexport OFF"
    assert testdev.com.get_written() == "SYST:DISP:UPD ON"

    # with timeout
    with pytest.raises(RTO1024Error):
        testdev.save_waveform_history("test", 1)

    # with "file not found"
    testdev.com.put_name(
        f"MMEMory:CATalog? '{testdev.config.waveforms_path}'",
        '142422792,147771314176,".,DIR,0","..,DIR,0","DavidGraber,DIR,0",'
        '"Myriam,DIR,0"',
    )
    testdev.com.stb = 32
    with pytest.raises(RTO1024Error):
        testdev.save_waveform_history("test", 1)
    testdev.com.stb = 0


def test_run_continuous_acquisition(testdev: RTO1024) -> None:
    testdev.run_continuous_acquisition()
    assert testdev.com.get_written() == "RUN"


def test_run_single_acquisition(testdev: RTO1024) -> None:
    testdev.run_single_acquisition()
    assert testdev.com.get_written() == "SINGle"


def test_stop_acquisition(testdev: RTO1024) -> None:
    testdev.stop_acquisition()
    assert testdev.com.get_written() == "STOP"


def test_prepare_ultra_segmentation(testdev: RTO1024) -> None:
    testdev.prepare_ultra_segmentation()


def test_save_configuration(testdev: RTO1024) -> None:
    testdev.save_configuration("test_configuration")
    assert (
        testdev.com.get_written()
        == "MMEMory:SAV 'C:\\Data\\DavidGraber\\01_settings\\test_configuration.dfl'"
    )


def test_load_configuration(testdev: RTO1024) -> None:
    testdev.load_configuration("test_configuration")
    assert (
        testdev.com.get_written()
        == "MMEMory:RCL 'C:\\Data\\DavidGraber\\01_settings\\test_configuration.dfl'"
    )


def test_get_timestamps(testdev: RTO1024) -> None:
    testdev.com.put_name("ACQuire:AVAilable?", "3")
    testdev.com.put_name("CHANnel:WAVeform:HISTory:TSRelative?", "-1.24")
    testdev.com.put_name("CHANnel:WAVeform:HISTory:TSRelative?", "-0.87")
    testdev.com.put_name("CHANnel:WAVeform:HISTory:TSRelative?", "0")
    ts = testdev.get_timestamps()
    assert ts == [-1.24, -0.87, 0]

    # with read errors
    testdev.com.put_name("ACQuire:AVAilable?", "3")
    testdev.com.put_name("CHANnel:WAVeform:HISTory:TSRelative?", "-1.24")
    testdev.com.put_name("CHANnel:WAVeform:HISTory:TSRelative?", "-1.24")
    testdev.com.put_name("CHANnel:WAVeform:HISTory:TSRelative?", "0")
    with pytest.raises(RTO1024Error):
        testdev.get_timestamps()


def test_activate_measurement(testdev: RTO1024) -> None:
    meas_n = 1
    source = "C1W2"
    meas = ["PDEL", "MIN", "MAX"]
    category = "AMPT"
    testdev.activate_measurements(meas_n, source, meas, category)
    assert testdev.com.get_written() == f"MEAS{meas_n}:ENAB ON"
    assert testdev.com.get_written() == f"MEAS{meas_n}:SOUR {source}"
    assert testdev.com.get_written() == f"MEAS{meas_n}:CAT {category}"
    if meas:
        assert testdev.com.get_written() == f"MEAS{meas_n}:MAIN {meas.pop(0)}"
    while meas:
        assert testdev.com.get_written() == f"MEAS{meas_n}:ADD {meas.pop(0)}, ON"


def test_read_measurement(testdev: RTO1024) -> None:
    meas_n = 3
    meas_name = "PDEL"
    testdev.com.put_name(f"MEAS{meas_n}:RES? {meas_name}", "0.12034")
    assert testdev.read_measurement(meas_n, meas_name) == 0.12034



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Cube package with implementation for system versions from 2019 on (new concept
with hard-PLC Siemens S7-1500 as CPU).
"""

from .base import (  # noqa: F401
    BaseCube,
    BaseCubeConfiguration,
    BaseCubeOpcUaCommunication,
    BaseCubeOpcUaCommunicationConfig,
)
from .constants import (  # noqa: F401
    AC_POWER_SETUPS,
    DC_POWER_SETUPS,
    STOP_SAFETY_STATUSES,
    DoorStatus,
    EarthingRodStatus,
    Polarity,
    PowerSetup,
    SafetyStatus,
)
from .errors import (  # noqa: F401
    CubeEarthingStickOperationError,
    CubeError,
    CubeRemoteControlError,
    CubeStatusChangeError,
    CubeStopError,
    PICubeTestParameterError,
)
from .picube import (  # noqa: F401
    PICube,
    PICubeConfiguration,
    PICubeOpcUaCommunication,
    PICubeOpcUaCommunicationConfig,
)
from .switches import SwitchOperatingStatus, SwitchOperation, SwitchStatus  # noqa: F401



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Main module containing the top level ExperimentManager class.
Inherit from this class to implement your own experiment functionality in another
project and it will help you start, stop and manage your devices.
"""

import logging
from enum import Enum

from hvl_ccb.dev import Device, DeviceSequenceMixin
from hvl_ccb.error import CCBError

logger = logging.getLogger(__name__)


class ExperimentStatus(Enum):
    """
    Enumeration for the experiment status
    """

    INITIALIZING = -1
    INITIALIZED = 0
    STARTING = 1
    RUNNING = 2
    FINISHING = 3
    FINISHED = 4
    ERROR = 5


class ExperimentError(CCBError):
    """
    Error to indicate that the current status of the experiment manager is on
    ERROR and thus no operations can be made until reset.
    """


class ExperimentManager(DeviceSequenceMixin):
    """
    Experiment Manager can start and stop communication protocols and devices.
    It provides methods to queue commands to
    devices and collect results.
    """

    def __new__(cls, *_args, **_kwargs) -> "ExperimentManager":
        """
        Construct a new `ExperimentManager` instance.

        :param args: `__init__` arguments
        :param kwargs: `__init__` keyword arguments
        :return: `ExperimentManager` with status set to `ExperimentStatus.INITIALIZING`.
        """
        obj = super().__new__(cls)
        # experiment manager status
        obj._status = ExperimentStatus.INITIALIZING
        return obj

    def __init__(self, devices: dict[str, Device]) -> None:
        """
        Initialize `ExperimentManager`. Takes a dictionary of instantiated devices
        and initializes the experiment status to `ExperimentStatus.INITIALIZED`.

        :param devices: Devices sequenced in the experiment.
        """
        super().__init__(devices)

        self._change_status(ExperimentStatus.INITIALIZED)

    def _change_status(self, new_status: ExperimentStatus) -> None:
        """
        Change the status of this experiment manager to a new value. Includes logging.

        :param new_status: is the Enum of the new experiment status.
        """

        if new_status is not self._status:
            log_msg = f"Experiment Status: {new_status} {new_status.name}"
            # new status is different, log
            if new_status is ExperimentStatus.ERROR:
                logger.error(log_msg)
            else:
                logger.info(log_msg)
            self._status: ExperimentStatus = new_status

    def run(self) -> None:
        """
        Start experimental setup, start all devices.
        """

        self._change_status(ExperimentStatus.STARTING)

        try:
            super().start()
        except CCBError as e:
            logger.exception("Error during stat-up of devices")
            self._change_status(ExperimentStatus.ERROR)
            raise ExperimentError from e

        self._change_status(ExperimentStatus.RUNNING)

    def finish(self) -> None:
        """
        Stop experimental setup, stop all devices.
        """

        self._change_status(ExperimentStatus.FINISHING)

        try:
            super().stop()
        except CCBError as e:
            logger.exception("Error during finishing the experiment", exc_info=e)
            self._change_status(ExperimentStatus.ERROR)
            raise ExperimentError from e
        else:
            self._change_status(ExperimentStatus.FINISHED)

    def start(self) -> None:
        """
        Alias for ExperimentManager.run()
        """
        self.run()

    def stop(self) -> None:
        """
        Alias for ExperimentManager.finish()
        """
        self.finish()

    @property
    def status(self) -> ExperimentStatus:
        """
        Get experiment status.

        :return: experiment status enum code.
        """

        return self._status

    def is_finished(self) -> bool:
        """
        Returns true, if the status of the experiment manager is `finished`.

        :return: True if finished, false otherwise
        """
        return self.status == ExperimentStatus.FINISHED

    def is_running(self) -> bool:
        """
        Returns true, if the status of the experiment manager is `running`.

        :return: True if running, false otherwise
        """
        return self.status == ExperimentStatus.RUNNING

    def is_error(self) -> bool:
        """
        Returns true, if the status of the experiment manager is `error`.

        :return: True if on error, false otherwise
        """
        return self.status == ExperimentStatus.ERROR

    def add_device(self, name: str, device: Device) -> None:
        """
        Add a new device to the manager. If the experiment is running, automatically
        start the device. If a device with this name already exists, raise an error.

        :param name: is the name of the device.
        :param device: is the instantiated Device object.
        :raise ExperimentError:
        """

        if self.status == ExperimentStatus.ERROR:
            err_msg = "Experiment is on ERROR, cannot add device"
            logger.error(err_msg)
            raise ExperimentError(err_msg)

        super().add_device(name, device)

        # check experiment status and start device if started
        if self.status == ExperimentStatus.RUNNING:
            logger.info(f"Experiment is already RUNNING, start Device {device}")
            try:
                device.start()
            except CCBError as e:
                logger.exception("Error at adding a new Device", exc_info=e)
                self._change_status(ExperimentStatus.ERROR)
                raise ExperimentError from e



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Tests for the newport sub-package.
"""

import logging

import pytest

from hvl_ccb.comm.serial import SerialCommunicationIOError
from hvl_ccb.dev.newport import (
    NewportConfigCommands,
    NewportControllerError,
    NewportMotorError,
    NewportMotorPowerSupplyWasCutError,
    NewportSerialCommunicationError,
    NewportSMC100PP,
    NewportSMC100PPConfig,
    NewportSMC100PPSerialCommunicationConfig,
    NewportUncertainPositionError,
)
from masked_comm.serial import NewportLoopSerialCommunication

logging.basicConfig(level=logging.ERROR)


@pytest.fixture(scope="module")
def com_config():
    return {
        "port": "loop://?logging=debug",
        "baudrate": 57600,
        "parity": NewportSMC100PPSerialCommunicationConfig.Parity.NONE,
        "stopbits": NewportSMC100PPSerialCommunicationConfig.Stopbits.ONE,
        "bytesize": NewportSMC100PPSerialCommunicationConfig.Bytesize.EIGHTBITS,
        "terminator": b"\r\n",
        "timeout": 0.05,
    }


@pytest.fixture(scope="module")
def dev_config():
    return {
        "address": 1,
        "user_position_offset": 10,
        "screw_scaling": 1,
        "exit_configuration_wait_sec": 0.01,
        "move_wait_sec": 0.01,
        "acceleration": 10,
        "backlash_compensation": 0,
        "hysteresis_compensation": 0.015,
        "micro_step_per_full_step_factor": 100,
        "motion_distance_per_full_step": 0.01,
        "home_search_type": 2,
        "jerk_time": 0.04,
        "home_search_velocity": 4,
        "home_search_timeout": 1.01,
        "home_search_polling_interval": 0.01,
        "peak_output_current_limit": 0.4,
        "rs485_address": 2,
        "negative_software_limit": -23.5,
        "positive_software_limit": 25,
        "velocity": 4,
        "base_velocity": 0,
        "stage_configuration": 3,
    }


@pytest.fixture
def started_newport(com_config, dev_config):
    com = NewportLoopSerialCommunication(com_config)
    com.open()

    mot = NewportSMC100PP(com, dev_config)
    motor_config = mot.config.motor_config

    # put answer to get_state query (NO_REF)
    state = NewportSMC100PP.StateMessages.NO_REF_FROM_RESET.value
    com.put_text(f"{dev_config['address']}TS0000{state}")
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")

    # put answer to get_configuration query (matching config values)
    com.put_text("1PW1")
    com.put_text("1ID hello")
    for param in motor_config:
        cmd = NewportConfigCommands(param).name
        val = motor_config[param]  # matching config values
        com.put_text(f"1{cmd}{val}")
        logging.debug(f"putting to com: 1{cmd}{val}")
    com.put_text("1PW0")
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")

    # put answer to initialize command
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")

    # put answer to get_state query in wait_until_motor_initialized (READY)
    state = NewportSMC100PP.StateMessages.READY_FROM_HOMING.value
    com.put_text(f"{dev_config['address']}TS0000{state}")
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")

    # put position as if just initialized
    mot.position = dev_config["user_position_offset"]

    with mot:
        while com.get_written() is not None:
            pass
        yield com, mot
        if com.is_open:
            # required for a correct dev stop
            # Note: this should be rather mot.is_started
            com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")


def test_com_read_text(com_config) -> None:
    com = NewportLoopSerialCommunication(com_config)
    com.open()
    # check that no error is raised
    com.put_text("regular text")
    assert com.read_text().strip() == "regular text"
    # check that NewportMotorPowerSupplyWasCutError is raised
    com.put_text("\x00regular text")
    with pytest.raises(NewportMotorPowerSupplyWasCutError):
        com.read_text()
    com.put_bytes(b"\xf8\x00")
    with pytest.raises(NewportMotorPowerSupplyWasCutError):
        com.read_text()


def test_com_check_for_error(com_config) -> None:
    com = NewportLoopSerialCommunication(com_config)
    com.open()
    # check that no error is raised
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    com.check_for_error(1)
    assert com.get_written() == "1TE"
    # check that NewportControllerError is raised
    com.put_text(f"1TE{com.ControllerErrors.CMD_NOT_ALLOWED.value}")
    with pytest.raises(NewportControllerError):
        com.check_for_error(1)
    assert com.get_written() == "1TE"
    # check that SerialCommunicationIOError is raised if the com is closed
    com.close()
    with pytest.raises(SerialCommunicationIOError):
        com.check_for_error(1)


def test_com_send_command(com_config) -> None:
    com = NewportLoopSerialCommunication(com_config)
    com.open()

    # check that the correct message is sent
    add = 3
    cmd = "command"
    param = 32.5
    com._send_command_without_checking_error(add, cmd, param)
    assert com.get_written() == f"{add}{cmd}{param}"

    # check that the command is sent an no error is signaled

    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    com.send_command(1, "command", 12)
    assert com.get_written() == "1command12"
    assert com.get_written() == "1TE"
    # check that NewportControllerError is raised
    com.put_text(f"1TE{com.ControllerErrors.CMD_NOT_ALLOWED.value}")
    with pytest.raises(NewportControllerError):
        com.send_command(1, "command")
    assert com.get_written() == "1command"
    assert com.get_written() == "1TE"
    # check that SerialCommunicationIOError is raised if the com is closed
    com.close()
    with pytest.raises(SerialCommunicationIOError):
        com.send_command(1, "command")


def test_com_query(com_config) -> None:
    com = NewportLoopSerialCommunication(com_config)
    com.open()
    # check that the correct answer is returned
    com.put_text("1query answer")
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    assert com.query(1, "query", "?") == "answer"
    assert com.get_written() == "1query?"
    assert com.get_written() == "1TE"
    # also without param
    com.put_text("1query answer")
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    assert com.query(1, "query") == "answer"
    # check that NewportSerialCommunicationError is raised if unexpected answer
    com.put_text("")
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    with pytest.raises(NewportSerialCommunicationError):
        com.query(1, "query")
    # check that NewportControllerError is raised
    com.put_text("wrong answer")
    com.put_text(f"1TE{com.ControllerErrors.CMD_NOT_ALLOWED.value}")
    with pytest.raises(NewportControllerError):
        com.query(1, "query")
    # check that SerialCommunicationIOError is raised if the com is closed
    com.close()
    with pytest.raises(SerialCommunicationIOError):
        com.query(1, "query")


def test_com_query_multiple(com_config) -> None:
    com = NewportLoopSerialCommunication(com_config)
    com.open()

    # normal case
    add = 3
    cmd = "ZT"
    prefixes = ["one", "two", "three"]
    values = [1, 2, 3]
    for p, v in zip(prefixes, values):
        com.put_text(f"{p}{v}")
    com.put_text(f"{add}TE{com.ControllerErrors.NO_ERROR.value}")
    assert com.query_multiple(add, cmd, prefixes) == [str(v) for v in values]

    # wrong prefix
    wrong_prefixes = ["one", "two", "four"]
    for p, v in zip(prefixes, values):
        com.put_text(f"{p}{v}")
    com.put_text(f"{add}TE{com.ControllerErrors.NO_ERROR.value}")
    with pytest.raises(NewportSerialCommunicationError):
        com.query_multiple(add, cmd, wrong_prefixes)

    # check that NewportControllerError is raised
    for p, v in zip(prefixes, values):
        com.put_text(f"{p}{v}")
    com.put_text(f"{add}TE{com.ControllerErrors.CMD_NOT_ALLOWED.value}")
    with pytest.raises(NewportControllerError):
        com.query_multiple(add, cmd, prefixes)

    # check that SerialCommunicationIOError is raised if the com is closed
    com.close()
    with pytest.raises(SerialCommunicationIOError):
        com.query_multiple(add, cmd, prefixes)


def test_dev_config(dev_config) -> None:
    # currently there are no non-default config values
    NewportSMC100PPConfig()

    config = NewportSMC100PPConfig(**dev_config)
    for key, value in dev_config.items():
        assert getattr(config, key) == value

    # test _motor_config update on force_value
    fieldname = "acceleration"
    value = dict(dev_config)[fieldname]
    # first call to motor_config creates cache
    assert config.motor_config[fieldname] == value
    value = value - 1
    config.force_value(fieldname, value)
    # check that cache is updated
    assert config.motor_config[fieldname] == value


@pytest.mark.parametrize(
    "wrong_config_dict",
    [
        {"address": -1},
        {"screw_scaling": -1},
        {"exit_configuration_wait_sec": -1},
        {"move_wait_sec": -1},
        {"acceleration": -1},
        {"backlash_compensation": -1},
        {"hysteresis_compensation": -1},
        {"micro_step_per_full_step_factor": -1},
        {"motion_distance_per_full_step": -1},
        {"home_search_type": -1},
        {"jerk_time": -1},
        {"home_search_velocity": -1},
        {"home_search_timeout": -1},
        {"home_search_polling_interval": -1},
        {"peak_output_current_limit": -1},
        {"rs485_address": -1},
        {"negative_software_limit": 1},
        {"positive_software_limit": -1},
        {"velocity": -1},
        {"base_velocity": -1},
        {"stage_configuration": -1},
    ],
)
def test_invalid_config_dict(dev_config, wrong_config_dict) -> None:
    invalid_config = dict(dev_config)
    invalid_config.update(wrong_config_dict)
    with pytest.raises(ValueError):
        NewportSMC100PPConfig(**invalid_config)


def test_newport_instantiation(com_config, dev_config) -> None:
    mot = NewportSMC100PP(com_config)
    assert mot is not None

    mot = NewportSMC100PP(com_config, dev_config)
    assert mot is not None


def test_newport_com_errors(com_config, dev_config) -> None:
    wrong_com_config = dict(com_config)
    wrong_com_config["port"] = "NOT A PORT"
    mot = NewportSMC100PP(wrong_com_config, dev_config)
    with pytest.raises(SerialCommunicationIOError):
        mot.start()

    mot = NewportSMC100PP(com_config, dev_config)
    with pytest.raises(SerialCommunicationIOError):
        mot.get_state()
    with pytest.raises(SerialCommunicationIOError):
        mot.get_motor_configuration()
    with pytest.raises(SerialCommunicationIOError):
        mot.go_to_configuration()
    with pytest.raises(SerialCommunicationIOError):
        mot.set_motor_configuration()
    with pytest.raises(SerialCommunicationIOError):
        mot.exit_configuration()
    with pytest.raises(SerialCommunicationIOError):
        mot.initialize()
    with pytest.raises(SerialCommunicationIOError):
        mot.wait_until_motor_initialized()
    with pytest.raises(SerialCommunicationIOError):
        mot.reset()
    with pytest.raises(SerialCommunicationIOError):
        mot.get_position()
    with pytest.raises(SerialCommunicationIOError):
        mot.move_to_absolute_position(10)
    with pytest.raises(SerialCommunicationIOError):
        mot.go_home()
    mot.position = 1
    with pytest.raises(SerialCommunicationIOError):
        mot.move_to_relative_position(1)
    with pytest.raises(SerialCommunicationIOError):
        mot.get_move_duration(10)
    with pytest.raises(SerialCommunicationIOError):
        mot.stop_motion()
    with pytest.raises(SerialCommunicationIOError):
        mot.get_acceleration()
    with pytest.raises(SerialCommunicationIOError):
        mot.set_acceleration(10)
    with pytest.raises(SerialCommunicationIOError):
        mot.get_controller_information()
    with pytest.raises(SerialCommunicationIOError):
        mot.get_positive_software_limit()
    with pytest.raises(SerialCommunicationIOError):
        mot.set_positive_software_limit(1)
    with pytest.raises(SerialCommunicationIOError):
        mot.get_negative_software_limit()
    with pytest.raises(SerialCommunicationIOError):
        mot.set_negative_software_limit(-1)


def test_start_and_initialize(com_config, dev_config) -> None:
    com = NewportLoopSerialCommunication(com_config)
    com.open()

    # case 1: config unchanged and in NO_REF state
    mot = NewportSMC100PP(com, dev_config)

    # answer to get_state
    state_message = NewportSMC100PP.StateMessages.NO_REF_FROM_RESET
    com.put_text(f"{dev_config['address']}TS0000" + state_message.value)
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")

    # answer to get_motor_configuration
    config = mot.config.motor_config
    com.put_text("1PW1")
    com.put_text("1ID hello")
    for param in config:
        cmd = NewportConfigCommands(param).name
        val = config[param]  # matching config values
        com.put_text(f"1{cmd}{val}")
        logging.debug(f"putting to com: 1{cmd}{val}")
    com.put_text("1PW0")
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")

    # answer to initialize
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    # answer to wait_until_motor_initialized
    state_message = NewportSMC100PP.StateMessages.READY_FROM_HOMING
    com.put_text(f"{dev_config['address']}TS0000" + state_message.value)
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")

    mot.start()
    assert mot.state == state_message.state


def test_start_and_configure(com_config, dev_config) -> None:
    com = NewportLoopSerialCommunication(com_config)
    com.open()

    # case 2: config changed and in READY state
    mot = NewportSMC100PP(com, dev_config)

    # answer to get_state
    state_message = NewportSMC100PP.StateMessages.READY_FROM_HOMING
    com.put_text(f"{dev_config['address']}TS0000" + state_message.value)
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")

    # answer to get_motor_configuration
    config = mot.config.motor_config
    com.put_text("1PW1")
    com.put_text("1ID hello")
    for param in config:
        cmd = NewportConfigCommands(param).name
        val = config[param] + 1  # changed config values
        com.put_text(f"1{cmd}{val}")
        logging.debug(f"putting to com: 1{cmd}{val}")
    com.put_text("1PW0")
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")

    #  answer to reset
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    com.put_text("I just did a reset and I feel talkative.")

    # answer to go_to_configuration
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    # answer to set_motor_configuration
    params = [p for p in config if not ("compensation" in p and config[p] == 0)]
    for _ in params:
        com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    # answer to exit_configuration
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")

    # answer to initialize
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    # answer to wait_until_motor_initialized
    state_message = NewportSMC100PP.StateMessages.READY_FROM_HOMING
    com.put_text(f"{dev_config['address']}TS0000" + state_message.value)
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")

    mot.start()
    assert mot.state == state_message.state


def test_stop(started_newport) -> None:
    com, mot = started_newport
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    mot.stop()


def test_get_state(started_newport) -> None:
    com, mot = started_newport

    state = NewportSMC100PP.StateMessages.MOVING.value
    com.put_text("1TS0000" + state)
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    mot.get_state()
    assert com.get_written() == "1TS"
    assert com.get_written() == "1TE"
    assert mot.state == mot.States.MOVING

    # test with another address
    state = NewportSMC100PP.StateMessages.READY_FROM_HOMING.value
    com.put_text("3TS0000" + state)
    com.put_text("3TE" + com.ControllerErrors.NO_ERROR.value)
    mot.get_state(3)
    assert com.get_written() == "3TS"
    assert com.get_written() == "3TE"
    assert mot.state == mot.States.READY

    # test when the motor reports an error
    state = NewportSMC100PP.StateMessages.DISABLE_FROM_READY.value
    com.put_text("1TS0013" + state)
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    with pytest.raises(NewportMotorError):
        mot.get_state()
    com.get_written()
    com.get_written()

    # test when an NewportMotorPowerSupplyWasCutError occurs and is skipped
    state = NewportSMC100PP.StateMessages.DISABLE_FROM_READY.value
    com.put_text("\x001TS0000" + state)
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    com.put_text("1TS0000" + state)
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    mot.get_state()
    assert com.get_written() == "1TS"
    assert com.get_written() == "1TE"
    assert com.get_written() == "1TS"
    assert com.get_written() == "1TE"
    assert mot.state == mot.States.DISABLE


def test_get_motor_configuration(started_newport) -> None:
    com, mot = started_newport

    com.put_text("1PW1")
    com.put_text("1ID hello")
    for param in mot.config.motor_config:
        cmd = NewportConfigCommands(param).name
        val = mot.config.motor_config[param]
        com.put_text(f"1{cmd}{val}")
        logging.debug(f"putting to com: 1{cmd}{val}")
    com.put_text("1PW0")
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    assert mot.get_motor_configuration() == mot.config.motor_config
    assert com.get_written() == "1ZT"
    assert com.get_written() == "1TE"


def test_go_to_configuration(started_newport) -> None:
    com, mot = started_newport
    # artificially put the controller in NOT REFERENCED state
    state = NewportSMC100PP.StateMessages.NO_REF_FROM_RESET.value
    com.put_text("1TS0000" + state)
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    mot.get_state()
    com.get_written()
    com.get_written()
    # start the test
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    mot.go_to_configuration()
    assert com.get_written() == "1PW1"
    assert com.get_written() == "1TE"
    assert mot.state == mot.States.CONFIG


def test_set_motor_configuration(started_newport) -> None:
    com, mot = started_newport

    config = mot.config.motor_config
    params = [p for p in config if not ("compensation" in p and config[p] == 0)]
    # normal case
    for _ in params:
        com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    mot.set_motor_configuration()
    for param in params:
        cmd = NewportConfigCommands(param).name
        val = config[param]
        assert com.get_written() == f"1{cmd}{val}"
        assert com.get_written() == "1TE"
    # controller error
    error = com.ControllerErrors.PARAM_MISSING_OR_INVALID.value
    com.put_text(f"1TE{error}")
    with pytest.raises(NewportControllerError):
        mot.set_motor_configuration()


def test_exit_configuration(started_newport) -> None:
    com, mot = started_newport
    # start the test
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    mot.exit_configuration()
    assert com.get_written() == "1PW0"
    assert com.get_written() == "1TE"
    assert mot.state == mot.States.NO_REF


def test_initialize(started_newport) -> None:
    com, mot = started_newport
    # start the test
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    mot.initialize()
    assert com.get_written() == "1OR"
    assert com.get_written() == "1TE"
    assert mot.state == mot.States.READY


def test_wait_until_motor_initialized(started_newport) -> None:
    com, mot = started_newport
    # normal case
    state = NewportSMC100PP.StateMessages.HOMING_FROM_RS232.value
    com.put_text("1TS0000" + state)
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    state = NewportSMC100PP.StateMessages.READY_FROM_HOMING.value
    com.put_text("1TS0000" + state)
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    mot.wait_until_motor_initialized()
    # unexpected final state or timeout
    state = NewportSMC100PP.StateMessages.HOMING_FROM_RS232.value
    com.put_text("1TS0000" + state)
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    state = NewportSMC100PP.StateMessages.DISABLE_FROM_MOVING.value
    com.put_text("1TS0000" + state)
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    with pytest.raises(NewportControllerError):
        mot.wait_until_motor_initialized()


def test_reset(started_newport) -> None:
    com, mot = started_newport
    # start the test
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    com.put_text("I just did a reset and I feel talkative.")
    mot.reset()
    assert com.get_written() == "1RS"
    assert com.get_written() == "1TE"
    assert mot.state == mot.States.NO_REF


def test_get_position(started_newport) -> None:
    com, mot = started_newport
    # start the test
    pos = 12
    com.put_text(f"1TP{pos - mot.config.user_position_offset}")
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    assert mot.get_position() == pos
    assert com.get_written() == "1TP"
    assert com.get_written() == "1TE"

    # check that an error is raised if the position is ambiguous
    com.put_text("1TP0")
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    # answer to the get_state() query
    state = NewportSMC100PP.StateMessages.NO_REF_FROM_RESET.value
    com.put_text("1TS0000" + state)
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    with pytest.raises(NewportUncertainPositionError):
        mot.get_position()


def test_move_to_absolute_position(started_newport) -> None:
    com, mot = started_newport
    # start the test
    pos = 3
    # answers to get_position()
    com.put_text(f"1TP{pos - mot.config.user_position_offset}")
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    # answers to absolute move command
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    mot.move_to_absolute_position(pos)
    assert com.get_written() == "1TP"
    assert com.get_written() == "1TE"
    assert com.get_written() == f"1PA{pos-mot.config.user_position_offset}"
    assert com.get_written() == "1TE"


def test_go_home(started_newport) -> None:
    com, mot = started_newport
    # start the test
    pos = 3
    # answers to get_position()
    com.put_text(f"1TP{pos - mot.config.user_position_offset}")
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    # answers to homing command
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    mot.go_home()
    assert com.get_written() == "1TP"
    assert com.get_written() == "1TE"
    assert com.get_written() == "1PA0"
    assert com.get_written() == "1TE"


def test_move_to_relative_position(started_newport) -> None:
    com, mot = started_newport
    # start the test
    pos = -3
    mot.position = 20
    # answers to relative move command
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    mot.move_to_relative_position(pos)
    assert com.get_written() == f"1PR{pos}"
    assert com.get_written() == "1TE"


def test_get_move_duration(started_newport) -> None:
    com, mot = started_newport
    dist = -3
    duration = 5.2
    mot.position = 0
    com.put_text(f"1PT{duration}")
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    assert mot.get_move_duration(dist) == duration
    assert com.get_written() == f"1PT{abs(dist)}"
    assert com.get_written() == "1TE"


def test_stop_motion(started_newport) -> None:
    com, mot = started_newport
    # test if stop_motion_works
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    mot.stop_motion()
    assert com.get_written() == "ST"
    assert com.get_written() == "1TE"
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    mot.stop_motion(1)
    assert com.get_written() == "1ST"
    assert com.get_written() == "1TE"


def test_get_acceleration(started_newport) -> None:
    com, mot = started_newport
    acc = 5
    com.put_text(f"1AC{acc}")
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    assert mot.get_acceleration() == acc
    assert com.get_written() == "1AC?"
    assert com.get_written() == "1TE"


def test_set_acceleration(started_newport) -> None:
    com, mot = started_newport
    acc = 5
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    mot.set_acceleration(acc)
    assert com.get_written() == f"1AC{acc}"
    assert com.get_written() == "1TE"


def test_get_controller_information(started_newport) -> None:
    com, mot = started_newport
    com.put_text("1VE SMC100PP")
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    assert mot.get_controller_information() == "SMC100PP"
    assert com.get_written() == "1VE?"
    assert com.get_written() == "1TE"


def test_get_positive_software_limit(started_newport) -> None:
    com, mot = started_newport
    lim = 5
    com.put_text(f"1SR{lim}")
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    assert mot.get_positive_software_limit() == lim
    assert com.get_written() == "1SR?"
    assert com.get_written() == "1TE"


def test_get_negative_software_limit(started_newport) -> None:
    com, mot = started_newport
    lim = -5
    com.put_text(f"1SL{lim}")
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    assert mot.get_negative_software_limit() == lim
    assert com.get_written() == "1SL?"
    assert com.get_written() == "1TE"


def test_set_positive_software_limit(started_newport) -> None:
    com, mot = started_newport
    lim = 5
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    mot.set_positive_software_limit(lim)
    assert com.get_written() == f"1SR{lim}"
    assert com.get_written() == "1TE"


def test_set_negative_software_limit(started_newport) -> None:
    com, mot = started_newport
    lim = -5
    com.put_text(f"1TE{com.ControllerErrors.NO_ERROR.value}")
    mot.set_negative_software_limit(lim)
    assert com.get_written() == f"1SL{lim}"
    assert com.get_written() == "1TE"



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Tests for map range in conversion
"""

import logging

import numpy as np
import pytest

from hvl_ccb.utils.conversion import MapBitAsymRange, MapBitSymRange, MapRanges


def test_external_or_internal_logger() -> None:
    logger = logging.getLogger(__name__)

    r = MapRanges((0, 10), (-5, 5), int, float)
    assert r.logger is not logger

    r = MapRanges((0, 10), (-5, 5), int, float, logger=logger)
    assert r.logger is logger


def test_bit_asym_range() -> None:
    bit10_10v = MapBitAsymRange(10, 10)
    assert bit10_10v.convert_to_bits(5) == 512
    assert round(bit10_10v.convert_to_number(512), 3) == 5.005


def test_bit_sym_range() -> None:
    bit10_10v = MapBitSymRange(5, 10)
    assert bit10_10v.convert_to_bits(5) == 1023
    assert round(bit10_10v.convert_to_number(20), 3) == -4.804
    assert bit10_10v.convert_to_bits(0) == 512


def test_check_dtype() -> None:
    assert MapRanges((0, 10), (-5, 5), float, float)
    assert MapRanges((0, 10), (-5, 5), float, np.float64)
    assert MapRanges((0, 10), (-5, 5), float, np.int_)

    with pytest.raises(TypeError):
        MapRanges((0, 10), (-5, 5), float, str)

    with pytest.raises(TypeError):
        MapRanges((0, 10), (-5, 5), float, np.complex128)

    with pytest.raises(TypeError):
        MapRanges((0, 10), (-5, 5), float, bool)


def test_check_range_type() -> None:
    assert MapRanges((0, 10), (-5, 5), float, float)

    with pytest.raises(TypeError):
        MapRanges((0, 10.5), (-5, 5), int, float)

    with pytest.raises(TypeError):
        MapRanges(("0", "10.5"), (-5, 5), int, float)


def test_check_value() -> None:
    r = MapRanges((0, 10), (-5, 5), int, float)
    assert r.convert_to_range1(0) == 5
    assert r.convert_to_range1((-5, 0, 5)) == (0, 5, 10)
    with pytest.raises(ValueError):
        r.convert_to_range2(11)

    with pytest.raises(TypeError):
        r.convert_to_range2("11")

    with pytest.raises(ValueError):
        r.convert_to_range1(11)

    with pytest.raises(TypeError):
        r.convert_to_range1("11")

    with pytest.raises(TypeError):
        r.convert_to_range2(11.1)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Ranges, RangeEnum for Fluke8845a device
"""

from typing import Literal

from hvl_ccb.utils.enum import RangeEnum


class DCVoltageRange(RangeEnum):
    """
    possible measurement ranges for DC voltage with unit volt
    """

    ONE_HUNDRED_MILLI_VOLT = 0.1
    ONE_VOLT = 1
    TEN_VOLT = 10
    HUNDRED_VOLT = 100
    ONE_THOUSAND_VOLT = 1000

    @classmethod
    def unit(cls) -> Literal["V"]:
        return "V"


class ACVoltageRange(RangeEnum):
    """
    possible measurement ranges for AC voltage with unit volt
    """

    ONE_HUNDRED_MILLI_VOLT = 0.1
    ONE_VOLT = 1
    TEN_VOLT = 10
    HUNDRED_VOLT = 100
    SEVEN_HUNDRED_FIFTY_VOLT = 750

    @classmethod
    def unit(cls) -> Literal["V"]:
        return "V"


class DCCurrentRange(RangeEnum):
    """
    possible measurement ranges for DC current with unit Ampere
    """

    ONE_HUNDRED_MICRO_AMPERE = 1e-4
    ONE_MILLI_AMPERE = 1e-3
    TEN_MILLI_AMPERE = 1e-2
    ONE_HUNDRED_MILLI_AMPERE = 0.1
    FOUR_HUNDRED_MILLI_AMPERE = 0.4
    ONE_AMPERE = 1
    THREE_AMPERE = 3
    TEN_AMPERE = 10

    @classmethod
    def unit(cls) -> Literal["A"]:
        return "A"


class ACCurrentRange(RangeEnum):
    """
    possible measurement ranges for AC current with unit Ampere
    """

    TEN_MILLI_AMPERE = 0.01
    ONE_HUNDRED_MILLI_AMPERE = 0.1
    FOUR_HUNDRED_MILLI_AMPERE = 0.4
    ONE_AMPERE = 1
    THREE_AMPERE = 3
    TEN_AMPERE = 10

    @classmethod
    def unit(cls) -> Literal["A"]:
        return "A"


class ResistanceRange(RangeEnum):
    """
    possible measurement ranges for resistance with unit Ohm
    """

    # Two-wire resistance (RES) and four-wire resistance (FRES)
    ONE_HUNDRED_OHM = 1e2
    ONE_THOUSAND_OHM = 1e3
    TEN_THOUSAND_OHM = 1e4
    ONE_HUNDRED_THOUSAND_OHM = 1e5
    ONE_MILLION_OHM = 1e6
    TEN_MILLION_OHM = 1e7
    ONE_HUNDRED_MILLION_OHM = 1e8

    @classmethod
    def unit(cls) -> Literal["Ω"]:
        return "Ω"


class ApertureRange(RangeEnum):
    """
    Page 46

    .. code-block:: rst

        Sets the gate time for the frequency/period function to the value
        10ms = 4 1/2 digits
        100ms = 5 1/2 digits
        1s = 6 1/2 digits
    """

    TEN_MILLI_SECOND = 0.01
    ONE_HUNDRED_MILLI_SECOND = 0.1
    ONE_SECOND = 1

    @classmethod
    def unit(cls) -> Literal["s"]:
        return "s"


class FilterRange(RangeEnum):
    """
    Page 47

    .. code-block:: rst

        Sets the appropriate filter for the frequency specified by <n>
        High pass filter
        For `VOLTAGE_AC`: <n> Hz to 300 kHz
        For `CURRENT_AC`: <n> Hz to 10 kHz
        parameters <n> = 3 slow filter
                        20 medium filter
                        200 fast filter
        For `CURRENT_AC` and `VOLTAGE_AC`
    """

    SLOW_FILTER = 3
    MEDIUM_FILTER = 20
    FAST_FILTER = 200

    @classmethod
    def unit(cls) -> Literal[""]:
        return ""



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#

from .map_range import MapBitAsymRange, MapBitSymRange, MapRanges  # noqa: F401
from .sensor import LEM4000S, LMT70A  # noqa: F401
from .unit import Pressure, Temperature  # noqa: F401



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Controller for the Pico Technology PT-104 temperature logger device.
The controller is written as a wrapper around Pico Technology driver for the PT-104
device.

This code is directly based on: https://github.com/trombastic/Pico_PT104/ .

Extra installation
~~~~~~~~~~~~~~~~~~

Pico Technology driver for the PT-104 device is available only on Windows and on Linux.

To use this PT-104 device wrapper:

1. install the :code:`hvl_ccb` package with a :code:`picotech` extra feature::

        $ pip install "hvl_ccb[picotech]"

   this will install the Python bindings for the library.

2. install the library

    * on Windows: download and install PicoSDK from https://www.picotech.com/downloads
      (choose "PicoLog Data Loggers" > "PT-104" > "Software");
    * on Linux:
        - for Ubuntu/Debian, install :code:`libusbpt104` from :code:`.deb` file found in
          https://labs.picotech.com/debian/pool/main/libu/libusbpt104/ (note: at the
          moment the PT-104 driver is not a part of the official :code:`picoscope`
          package; cf.
          https://www.picotech.com/support/topic40626.html );
        - for any other supported Linux distribution, follow instructions to install
          the "USB PT-104 devices" drivers in https://www.picotech.com/downloads/linux ;

"""

import sys

if sys.platform == "darwin":
    import warnings

    warnings.warn("\n\n  PicoSDK is not available for Darwin OSs\n", stacklevel=2)
else:
    try:
        from .picotech_pt104 import (  # noqa: F401
            Pt104,
            Pt104ChannelConfig,
            Pt104CommunicationType,
            Pt104DeviceConfig,
        )
    except (ImportError, ModuleNotFoundError):
        import warnings

        warnings.warn(
            "\n\n  "
            "To use PicoTech PT-104 device controller or related utilities install"
            "\n  "
            "PicoSDK and install the hvl_ccb library with a 'picotech' extra feature:"
            "\n\n  "
            "    $ pip install hvl_ccb[picotech]"
            "\n\n",
            stacklevel=2,
        )



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Constants of FuG
"""

from aenum import IntEnum

from hvl_ccb.utils.enum import NameEnum


class FuGDigitalVal(IntEnum):
    """
    Enum for digital values understood by FuG
    """

    OFF = 0
    ON = 1
    YES = 1
    NO = 0


class FuGRampModes(IntEnum):
    """
    Enum for the different ramp modes
    """

    IMMEDIATELY = 0
    """Standard mode: no ramp"""
    FOLLOWRAMP = 1
    """Follow the ramp up- and downwards"""
    RAMPUPWARDS = 2
    """Follow the ramp only upwards, downwards immediately"""
    SPECIALRAMPUPWARDS = 3
    """Follow a special ramp function only upwards"""
    ONLYUPWARDSOFFTOZERO = 4
    """Follow the ramp up- and downwards, if output is OFF set value is zero"""


class FuGReadbackChannels(IntEnum):
    """
    Enum for the readback channels for Trigger-on-Talk
    """

    VOLTAGE = 0
    CURRENT = 1
    STATUSBYTE = 2
    RATEDVOLTAGE = 3
    RATEDCURRENT = 4
    FIRMWARE = 5
    SN = 6


class FuGMonitorModes(IntEnum):
    """
    Enum for the different ADC modes
    """

    T256US = 0
    """14 bit + sign, 256 us integration time"""
    T1MS = 1
    """15 bit + sign, 1 ms integration time"""
    T4MS = 2
    """15 bit + sign, 4 ms integration time"""
    T20MS = 3
    """17 bit + sign, 20 ms integration time"""
    T40MS = 4
    """17 bit + sign, 40 ms integration time"""
    T80MS = 5
    """typ. 18 bit + sign, 80 ms integration time"""
    T200MS = 6
    """typ. 19 bit + sign, 200 ms integration time"""
    T800MS = 7
    """typ. 20 bit + sign, 800 ms integration time"""


class FuGPolarities(IntEnum):
    """
    Enum for the different polarities
    """

    POSITIVE = 0
    NEGATIVE = 1


class FuGTerminators(IntEnum):
    """
    Enum for the terminators of the command strings
    """

    CRLF = 0
    LFCR = 1
    LF = 2
    CR = 3


class FuGProbusIVCommands(NameEnum, init="command input_type"):  # type: ignore[call-arg]
    """
    Enum for the command of the (older) "Probus IV" interface.
    The `input_type` defines the type of a value that can be written.
    """

    ID = "*IDN?", None
    RESET = "=", None
    OUTPUT = "F", (FuGDigitalVal, int)
    VOLTAGE = "U", (int, float)
    CURRENT = "I", (int, float)
    READBACKCHANNEL = "N", (FuGReadbackChannels, int)
    QUERY = "?", None
    ADMODE = "S", (FuGMonitorModes, int)
    POLARITY = "P", (FuGPolarities, int)
    XOUTPUTS = "R", int
    """TODO: the possible values are limited to 0..13"""
    EXECUTEONX = "G", (FuGDigitalVal, int)
    """Wait for "X" to execute pending commands"""
    EXECUTE = "X", None
    TERMINATOR = "Y", (FuGTerminators, int)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Python module for the Rhode & Schwarz RTO 1024 oscilloscope.
The communication to the device is through VISA, type TCPIP / INSTR.
"""

from .rs_rto1024 import (  # noqa: F401
    RTO1024,
    RTO1024Config,
    RTO1024Error,
    RTO1024VisaCommunication,
    RTO1024VisaCommunicationConfig,
)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Tests for configuration.py: Mixin and class decorator configdataclass
"""

from dataclasses import field
from pathlib import Path
from typing import Union

import pytest

from hvl_ccb.configuration import ConfigurationMixin, EmptyConfig, configdataclass


@configdataclass
class GenericTypeHintsConfiguration:
    field1: list
    field2: list[int]
    field3: list


@configdataclass
class MyConfiguration:
    field1: int
    field2: str = "hello"

    def clean_values(self) -> None:
        if self.field1 == 0:
            msg = "field1 is not allowed to be 0"
            raise ValueError(msg)

        if self.field2 == "":
            self.force_value("field2", "EMPTY")

    def post_force_value(self, fieldname, _value) -> None:
        self.post_force_value.applied_to.append(fieldname)

    post_force_value.applied_to = []  # noqa: RUF012


@configdataclass
class MyConfigurationDefaults:
    field1: int
    field2: list
    field3: int = 1
    field4: list = field(default_factory=list)


@configdataclass
class SimpleConfiguration:
    field1: str = "field1"


class MyClassHasConfiguration(ConfigurationMixin):
    @staticmethod
    def config_cls():
        return MyConfiguration


# TESTS
def test_generic_type_hints() -> None:
    GenericTypeHintsConfiguration(["a"], [1], ["a"])
    with pytest.raises(TypeError):
        GenericTypeHintsConfiguration(("a",), [1], ["a"])
    with pytest.raises(TypeError):
        GenericTypeHintsConfiguration(["a"], [1], ("a",))
    with pytest.raises(TypeError):
        GenericTypeHintsConfiguration(["a"], ["a"], ["a"])


def test_inheritance() -> None:
    @configdataclass
    class SubMyConfiguration(MyConfiguration):
        field1: int = 3
        field3: int = 5

    assert SubMyConfiguration.optional_defaults() == {
        "field1": 3,
        "field2": "hello",
        "field3": 5,
    }

    SubMyConfiguration()

    with pytest.raises(AttributeError):

        @configdataclass
        class SubFailingConfiguration(MyConfiguration):
            def keys(self):
                pass


def test_configdataclass() -> None:
    SimpleConfiguration()

    # test own clean_values implementation
    with pytest.raises(ValueError):
        MyConfiguration(0)

    # test force_value
    config = MyConfiguration(1, field2="")
    assert config.field2 == "EMPTY"
    assert config.post_force_value.applied_to == ["field2"]

    # test convenience functions
    assert MyConfiguration.optional_defaults() == {"field2": "hello"}
    assert MyConfiguration.required_keys() == ["field1"]
    assert MyConfiguration.keys() == ["field1", "field2"]

    # typing is enforced
    with pytest.raises(TypeError):
        MyConfiguration("test")


def test_default_fields() -> None:
    assert MyConfigurationDefaults.required_keys() == ["field1", "field2"]
    assert MyConfigurationDefaults.optional_defaults() == {"field3": 1, "field4": []}


def test_configdataclass_with_generic() -> None:
    @configdataclass
    class ConfigurationWithGeneric:
        number: Union[int, float]

    with pytest.raises(TypeError):
        ConfigurationWithGeneric("1")
    assert ConfigurationWithGeneric(1).number == 1
    assert ConfigurationWithGeneric(1.0).number == 1.0


def test_configdataclass_fromdict() -> None:
    test_config = {"field1": 1, "field2": "otherstring"}

    my_class = MyClassHasConfiguration(test_config)
    assert my_class.config.field1 == 1
    assert my_class.config.field2 == "otherstring"

    test_config_2 = {"field3": 3, "field4": 4}

    # superfluous fields are not allowed
    with pytest.raises(TypeError):
        MyClassHasConfiguration(test_config_2)

    # non-configdataclasses are not allowed
    class WrongDefaultConfigDataclass(ConfigurationMixin):
        @staticmethod
        def config_cls():
            class NotAConfigDataclass:
                field1: int = 0

            return NotAConfigDataclass

    with pytest.raises(TypeError):
        WrongDefaultConfigDataclass({"field1": 1})


def test_empty_config() -> None:
    config = EmptyConfig()
    assert not config.required_keys()
    assert not config.keys()


def test_configuration_mixin() -> None:
    test_config = MyConfiguration(1)
    my_class = MyClassHasConfiguration(test_config)
    assert my_class.config == test_config

    with pytest.raises(TypeError):
        MyClassHasConfiguration(object)


def test_json_save_load() -> None:
    """
    Test the JSON configuration save and load feature.
    """

    c = MyClassHasConfiguration({"field1": 1})
    c.configuration_save_json("test.json")

    # create new protocol from that JSON configuration
    d = MyClassHasConfiguration.from_json("test.json")

    # clean up, delete file
    Path("test.json").unlink()

    assert c.config == d.config


def test_unfrozen() -> None:
    @configdataclass(frozen=False)
    class Test:
        field1: int = 3

    test = Test()
    assert test.field1 == 3
    test.field1 = 5
    assert test.field1 == 5

    # attention: typing is not enforced!
    test.field1 = "bla"
    assert test.field1 == "bla"



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Mixin class for Heinzinger's deprecated methods.Raise DeprecationWarning, will be
removed in the next release
"""

import logging

logger = logging.getLogger(__name__)


class DeprecatedHeinzingerMixin:
    @staticmethod
    def output_on() -> None:
        msg = (
            "output_on will be deprecated in the next release; "
            "use property instead: device.output = True"
        )
        logger.error(msg)
        raise DeprecationWarning(msg)

    @staticmethod
    def output_off() -> None:
        msg = (
            "output_off will be deprecated in the next release; "
            "use property instead: device.output = False"
        )
        logger.error(msg)
        raise DeprecationWarning(msg)

    @staticmethod
    def get_number_of_recordings() -> None:
        msg = (
            "get_number_of_recordings will be deprecated in the next release; "
            "use property instead: device.number_of_recordings"
        )
        logger.error(msg)
        raise DeprecationWarning(msg)

    @staticmethod
    def set_number_of_recordings() -> None:
        msg = (
            "set_number_of_recordings will be deprecated in the next release; "
            "use property instead: device.number_of_recordings = value"
        )
        logger.error(msg)
        raise DeprecationWarning(msg)

    @staticmethod
    def measure_voltage() -> None:
        msg = (
            "measure_voltage will be deprecated in the next release; "
            "use property instead: device.voltage"
        )
        logger.error(msg)
        raise DeprecationWarning(msg)

    @staticmethod
    def measure_current() -> None:
        msg = (
            "measure_current will be deprecated in the next release; "
            "use property instead: device.current"
        )
        logger.error(msg)
        raise DeprecationWarning(msg)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Device class for a SST Luminox Oxygen sensor. This device can measure the oxygen
concentration between 0 % and 25 %.

Furthermore, it measures the barometric pressure and internal temperature.
The device supports two operating modes: in streaming mode the device measures all
parameters every second, in polling mode the device measures only after a query.

Technical specification and documentation for the device can be found a the
manufacturer's page:
https://www.sstsensing.com/product/luminox-optical-oxygen-sensors-2/
"""

from .sst_luminox import (  # noqa: F401
    Luminox,
    LuminoxConfig,
    LuminoxMeasurementType,
    LuminoxMeasurementTypeError,
    LuminoxOutputMode,
    LuminoxOutputModeError,
    LuminoxSerialCommunication,
    LuminoxSerialCommunicationConfig,
)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Communication protocol for LabJack using the LJM Library.
Originally developed and tested for LabJack T7-PRO.

Makes use of the LabJack LJM Library Python wrapper.
This wrapper needs an installation of the LJM Library for Windows, Mac OS X or Linux.
Go to:
https://labjack.com/support/software/installers/ljm
and
https://labjack.com/support/software/examples/ljm/python
"""

import logging
from collections.abc import Sequence
from typing import Union

from labjack import ljm

from hvl_ccb._dev import labjack
from hvl_ccb.comm import CommunicationError, CommunicationProtocol
from hvl_ccb.configuration import configdataclass
from hvl_ccb.utils.enum import AutoNumberNameEnum
from hvl_ccb.utils.typing import Number

logger = logging.getLogger(__name__)


class LJMCommunicationError(CommunicationError):
    """
    Errors coming from LJMCommunication.
    """


@configdataclass
class LJMCommunicationConfig:
    """
    Configuration dataclass for :class:`LJMCommunication`.
    """

    DeviceType = labjack.DeviceType

    #: Can be either string 'ANY', 'T7_PRO', 'T7', 'T4', or of enum :class:`DeviceType`.
    device_type: Union[str, labjack.DeviceType] = "ANY"

    class ConnectionType(AutoNumberNameEnum):
        """
        LabJack connection type.
        """

        ANY = ()
        USB = ()
        TCP = ()
        ETHERNET = ()
        WIFI = ()

    #: Can be either string or of enum :class:`ConnectionType`.
    connection_type: Union[str, ConnectionType] = "ANY"

    identifier: str = "ANY"
    """
    The identifier specifies information for the connection to be used. This can
    be an IP address, serial number, or device name. See the LabJack docs (
    https://labjack.com/support/software/api/ljm/function-reference/ljmopens/\
identifier-parameter) for more information.
    """

    def clean_values(self) -> None:
        """
        Performs value checks on device_type and connection_type.
        """
        if not isinstance(self.device_type, self.DeviceType):
            self.force_value("device_type", self.DeviceType(self.device_type))  # type: ignore[attr-defined]

        if not isinstance(self.connection_type, self.ConnectionType):
            self.force_value(  # type: ignore[attr-defined]
                "connection_type", self.ConnectionType(self.connection_type)
            )


class LJMCommunication(CommunicationProtocol):
    """
    Communication protocol implementing the LabJack LJM Library Python wrapper.
    """

    def __init__(self, configuration) -> None:
        """
        Constructor for LJMCommunication.
        """
        super().__init__(configuration)

        # reference to the ctypes handle
        self._handle = None

    @staticmethod
    def config_cls():
        return LJMCommunicationConfig

    def open(self) -> None:
        """
        Open the communication port.
        """

        logger.info("Open connection")

        # open connection and store handle
        # may throw 1227 LJME_DEVICE_NOT_FOUND if device is not found
        try:
            with self.access_lock:
                self._handle = ljm.openS(
                    self.config.device_type.type_str,
                    str(self.config.connection_type),
                    str(self.config.identifier),
                )
        except ljm.LJMError as e:
            logger.exception("Error of LabJack LJM", exc_info=e)
            # only catch "1229 LJME_DEVICE_ALREADY_OPEN", never observed
            if e.errorCode != 1229:
                raise LJMCommunicationError from e

    def close(self) -> None:
        """
        Close the communication port.
        """

        logger.info("Closing connection")

        try:
            with self.access_lock:
                ljm.close(self._handle)
        except ljm.LJMError as e:
            logger.exception("Error of LabJack LJM", exc_info=e)
            # only catch "1224 LJME_DEVICE_NOT_OPEN", thrown on invalid handle
            if e.errorCode != 1224:
                raise LJMCommunicationError from e
        self._handle = None

    @property
    def is_open(self) -> bool:
        """
        Flag indicating if the communication port is open.

        :return: `True` if the port is open, otherwise `False`
        """
        # getHandleInfo does not work with LJM DEMO_MODE - consider it always opened
        # if only set
        if str(self._handle) == labjack.constants.DEMO_MODE:
            return True

        try:
            ljm.getHandleInfo(self._handle)
        except ljm.LJMError as e:
            if e.errorCode == 1224:  # "1224 LJME_DEVICE_NOT_OPEN"
                logger.warning("Error of LabJack LJM: '1224 LJME_DEVICE_NOT_OPEN'")
                return False
            logger.exception("Error of LabJack LJM", exc_info=e)
            raise LJMCommunicationError from e
        return True

    def __del__(self) -> None:
        """
        Finalizer, closes port
        """

        self.close()

    @staticmethod
    def _cast_read_value(
        name: str,
        val: Number,
        return_num_type: type[Number] = float,
        # see: https://github.com/python/mypy/issues/3186
    ) -> Number:
        """
        Cast a read value to a numeric type, performing some extra cast validity checks.

        :param name: name of the read value, only for error reporting
        :param val: value to cast
        :param return_num_type: optional numeric type specification for return values;
            by default `float`
        :return: input value `val` casted to `return_num_type`
        :raises TypeError: if read value of type not compatible with `return_num_type`
        """
        # Note: the underlying library returns already `float` (or
        # `ctypes.c_double`?); but defensively cast again via `str`:
        # 1) in case the underlying lib behaviour changes, and
        # 2) to raise `TypeError` when got non integer `float` value and expecting
        #    `int` value
        invalid_value_type = False
        try:
            fval = float(str(val))
            if issubclass(return_num_type, int) and not fval.is_integer():
                invalid_value_type = True
            else:
                ret = return_num_type(fval)
        except ValueError:
            invalid_value_type = True
        if invalid_value_type:
            msg = (
                f"Expected {return_num_type} value for '{name}' "
                f"name, got {type(val)} value of {val}"
            )
            logger.exception(msg)
            raise TypeError(msg)
        return ret

    def read_name(
        self,
        *names: str,
        return_num_type: type[Number] = float,
        # see: https://github.com/python/mypy/issues/3186
    ) -> Union[Number, Sequence[Number]]:
        """
        Read one or more input numeric values by name.

        :param names: one or more names to read out from the LabJack
        :param return_num_type: optional numeric type specification for return values;
            by default `float`.
        :return: answer of the LabJack, either single number or multiple numbers in a
            sequence, respectively, when one or multiple names to read were given
        :raises TypeError: if read value of type not compatible with `return_num_type`
        """

        # Errors that can be returned here:
        # 1224 LJME_DEVICE_NOT_OPEN if the device is not open
        # 1239 LJME_DEVICE_RECONNECT_FAILED if the device was opened, but connection
        #   lost

        with self.access_lock:
            try:
                if len(names) == 1:
                    ret = ljm.eReadName(self._handle, names[0])
                    ret = self._cast_read_value(
                        names[0], ret, return_num_type=return_num_type
                    )
                else:
                    ret = ljm.eReadNames(self._handle, len(names), names)
                    for i, (iname, iret) in enumerate(zip(names, ret)):
                        ret[i] = self._cast_read_value(
                            iname, iret, return_num_type=return_num_type
                        )
            except ljm.LJMError as e:
                logger.exception("Error of LabJack LJM", exc_info=e)
                raise LJMCommunicationError from e

        return ret

    def write_name(self, name: str, value: Number) -> None:
        """
        Write one value to a named output.

        :param name: String or with name of LabJack IO
        :param value: is the value to write to the named IO port
        """

        with self.access_lock:
            try:
                ljm.eWriteName(self._handle, name, value)
            except ljm.LJMError as e:
                logger.exception("Error of LabJack LJM", exc_info=e)
                raise LJMCommunicationError from e

    def write_names(self, name_value_dict: dict[str, Number]) -> None:
        """
        Write more than one value at once to named outputs.

        :param name_value_dict: is a dictionary with string names of LabJack IO as keys
            and corresponding numeric values
        """
        names = list(name_value_dict.keys())
        values = list(name_value_dict.values())
        with self.access_lock:
            try:
                ljm.eWriteNames(self._handle, len(names), names, values)
            except ljm.LJMError as e:
                logger.exception("Error of LabJack LJM", exc_info=e)
                raise LJMCommunicationError from e



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Supports of the different "Cubes".
"""

import logging
from typing import TYPE_CHECKING

from hvl_ccb.utils.validation import validate_bool

if TYPE_CHECKING:
    from . import BaseCube  # pragma: no cover

logger = logging.getLogger(__name__)


class _SupportPort:
    """
    class to collect outputs and inputs of support ports.
    """

    _PORTS: tuple = (1, 2, 3, 4, 5, 6)
    _CONTACTS: tuple = (1, 2)
    _IOS: tuple = ("Q", "I")

    def __init__(self, handle, number: int) -> None:
        self._handle: BaseCube = handle
        self._number: int = number

    def _cmd(self, io: str, contact: int) -> str:
        return f'"{io}x_Allg_Support{self._number}_{contact}"'

    @property
    def output_1(self) -> bool:
        """
        Output 1 of support port.

        :return: `True` if high, `False` if low
        """
        value = bool(self._handle.read(self._cmd("Q", 1)))
        self._log_msg(value, "Output", 1, "")
        return value

    @output_1.setter
    def output_1(self, value: bool) -> None:
        """
        Output 1 of support port.

        :param value: `True` for high output, `False` for low
        """
        validate_bool("state", value, logger)
        self._handle.write(self._cmd("Q", 1), value)
        self._log_msg(value, "Output", 1, "set to ")

    @property
    def output_2(self) -> bool:
        """
        Output 2 of support port.

        :return: `True` if high, `False` if low
        """
        value = bool(self._handle.read(self._cmd("Q", 2)))
        self._log_msg(value, "Output", 2, "")
        return value

    @output_2.setter
    def output_2(self, value: bool) -> None:
        """
        Output 2 of support port.

        :param value: `True` for high output, `False` for low
        """
        validate_bool("state", value, logger)
        self._handle.write(self._cmd("Q", 2), value)
        self._log_msg(value, "Output", 2, "set to ")

    @property
    def input_1(self) -> bool:
        """
        Input 1 of support port.

        :return: `True` if high, `False` if low
        """
        value = bool(self._handle.read(self._cmd("I", 1)))
        self._log_msg(value, "Input", 1, "")
        return value

    @property
    def input_2(self) -> bool:
        """
        Input 2 of support port

        :return: `True` if high, `False` if low
        """
        value = bool(self._handle.read(self._cmd("I", 2)))
        self._log_msg(value, "Input", 2, "")
        return value

    def _log_msg(self, value, name, contact, action) -> None:
        value_str = "HIGH" if value else "LOW"
        logger.info(
            f"Support {name} Port {self._number} Contact {contact} "
            f"is {action}{value_str}"
        )



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Tests for the Communication Protocol package comm.modbus_tcp.
"""

import pytest

from hvl_ccb.comm.modbus_tcp import (
    ModbusTcpCommunication,
    ModbusTcpCommunicationConfig,
    ModbusTcpConnectionFailedError,
)


@pytest.fixture(scope="module")
def testconfig():
    return ModbusTcpCommunicationConfig(
        host="127.0.0.1",
        unit=0,
    )


def test_modbus_tcp_conifg() -> None:
    with pytest.raises(ValueError):
        ModbusTcpCommunicationConfig(host="127..1", unit=0)


def test_modbus_tcp_connect(testconfig) -> None:
    mbtcp = ModbusTcpCommunication(testconfig)

    assert mbtcp is not None

    with pytest.raises(ModbusTcpConnectionFailedError):
        mbtcp.open()

    # test write
    with pytest.raises(ModbusTcpConnectionFailedError):
        mbtcp.write_registers(address=1234, values=5678)

    # test read holding registers
    with pytest.raises(ModbusTcpConnectionFailedError):
        mbtcp.read_holding_registers(address=9012, count=2)

    # test read input registers
    with pytest.raises(ModbusTcpConnectionFailedError):
        mbtcp.read_input_registers(address=3456, count=2)

    mbtcp.close()



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Tests for the hvl_ccb.dev.heinzinger sub-package.
"""

import logging

import pytest

from hvl_ccb.comm.serial import SerialCommunicationIOError
from hvl_ccb.dev.heinzinger.base import HeinzingerSerialCommunicationConfig
from hvl_ccb.dev.heinzinger.constants import (
    HeinzingerDeviceNotRecognizedError,
    HeinzingerSetValueError,
    RecordingsEnum,
)
from hvl_ccb.dev.heinzinger.device import Heinzinger, HeinzingerConfig
from masked_comm.serial import HeinzingerLoopSerialCommunication

logging.basicConfig(level=logging.ERROR)


@pytest.fixture(scope="module")
def com_config():
    return {
        "port": "loop://?logging=debug",
        "baudrate": 9600,
        "parity": HeinzingerSerialCommunicationConfig.Parity.NONE,
        "stopbits": HeinzingerSerialCommunicationConfig.Stopbits.ONE,
        "bytesize": HeinzingerSerialCommunicationConfig.Bytesize.EIGHTBITS,
        "terminator": b"\r\n",
        "timeout": 3,
        "wait_sec_read_text_nonempty": 0.01,
        "default_n_attempts_read_text_nonempty": 5,
    }


@pytest.fixture(scope="module")
def dev_config():
    return {
        "default_number_of_recordings": 16,
        "number_of_decimals": 6,
        "wait_sec_stop_commands": 0.01,
    }


class ConcreteHeinzinger(Heinzinger):
    def start(self) -> None:
        super().start()


@pytest.fixture
def started_heinzinger_device(com_config, dev_config):
    serial_port = HeinzingerLoopSerialCommunication(com_config)
    serial_port.open()
    serial_port.put_text("my_interface_version")
    serial_port.put_text("PNChp 60000-1neg 354211082")
    with Heinzinger(serial_port, dev_config) as heinzinger:
        while serial_port.get_written() is not None:
            pass
        yield serial_port, heinzinger


@pytest.fixture
def start_heinzinger_device(com_config, dev_config):
    def _start_heinzinger_device(serial_number):
        serial_port = HeinzingerLoopSerialCommunication(com_config)
        serial_port.open()
        serial_port.put_text("my_interface_version")
        serial_port.put_text(serial_number)

        def started_heinzinger():
            with Heinzinger(serial_port, dev_config) as heinzinger:
                while serial_port.get_written() is not None:
                    pass
                yield serial_port, heinzinger

        return started_heinzinger()

    return _start_heinzinger_device


def test_heinzinger_number_of_recordings(started_heinzinger_device) -> None:
    com, heinzinger = started_heinzinger_device
    # set a new value
    heinzinger.number_of_recordings = 4
    assert com.get_written() == "AVER 4"

    com.put_text("4")
    assert heinzinger.number_of_recordings == 4
    assert com.get_written() == "AVER?"

    # assigning integer not amongst accepted values
    # RangeEnum will pick a suitable value automatically, set 6 will get 8
    heinzinger.number_of_recordings = 6
    assert com.get_written() == "AVER 8"
    com.put_text("0.0")  # when device stop, voltage is set to 0


def test_com_config(com_config) -> None:
    config = HeinzingerSerialCommunicationConfig(**com_config)
    for key, value in com_config.items():
        assert getattr(config, key) == value


def test_dev_config(dev_config) -> None:
    # currently there are no non-default config values
    HeinzingerConfig()

    config = HeinzingerConfig(**dev_config)
    for key, value in dev_config.items():
        assert getattr(config, key) == value


@pytest.mark.parametrize(
    "wrong_config_dict",
    [
        {"number_of_decimals": 0},
        {"number_of_decimals": 11},
        {"wait_sec_stop_commands": 0},
        {"number_of_decimals": 6, "wait_sec_stop_commands": -1},
    ],
)
def test_invalid_config_dict(dev_config, wrong_config_dict) -> None:
    invalid_config = dict(dev_config)
    invalid_config.update(wrong_config_dict)
    with pytest.raises(ValueError):
        HeinzingerConfig(**invalid_config)


def test_heinzinger_instantiation(com_config, dev_config) -> None:
    heinzinger = ConcreteHeinzinger(com_config)
    assert heinzinger is not None

    heinzinger = ConcreteHeinzinger(com_config, dev_config)
    assert heinzinger is not None


def test_heinzinger_stop(started_heinzinger_device) -> None:
    com, heinzinger = started_heinzinger_device
    com.put_text("0.0")
    heinzinger.stop()
    assert com.get_written() == "VOLT 0.000000"
    assert com.get_written() == "VOLT?"
    assert com.get_written() == "OUTP OFF"

    # check that a second stop() works
    heinzinger.stop()
    assert not com.get_written()


def test_heinzinger_reset_interface(started_heinzinger_device) -> None:
    com, heinzinger = started_heinzinger_device
    heinzinger.reset_interface()
    assert com.get_written() == "*RST"
    com.put_text("0.0")  # when device stop, voltage is set to 0


def test_heinzinger_output_status_property(started_heinzinger_device) -> None:
    com, heinzinger = started_heinzinger_device
    assert heinzinger.output is None
    heinzinger.output = True
    assert heinzinger.output is True
    heinzinger.output = False
    assert heinzinger.output is False
    com.put_text("0.0")  # when device stop, voltage is set to 0


def test_heinzinger_com_error(com_config, dev_config) -> None:
    wrong_config = dict(com_config)
    wrong_config["port"] = "NOT A PORT"
    heinzinger = ConcreteHeinzinger(wrong_config, dev_config)
    assert not heinzinger.com.is_open

    with pytest.raises(SerialCommunicationIOError):
        heinzinger.start()

    heinzinger = ConcreteHeinzinger(com_config, dev_config)
    assert not heinzinger.com.is_open
    heinzinger._max_current_hardware = 1.5
    heinzinger._max_voltage_hardware = 1.5
    heinzinger._max_current = 1.5
    heinzinger._max_voltage = 1.5
    heinzinger._voltage_multiplier = 1.0
    heinzinger._current_multiplier = 1.0
    with pytest.raises(SerialCommunicationIOError):
        heinzinger.reset_interface()
    with pytest.raises(SerialCommunicationIOError):
        heinzinger.get_interface_version()
    with pytest.raises(SerialCommunicationIOError):
        heinzinger.get_serial_number()
    with pytest.raises(SerialCommunicationIOError):
        assert heinzinger.number_of_recordings
    with pytest.raises(SerialCommunicationIOError):
        heinzinger.number_of_recordings = RecordingsEnum.EIGHT
    with pytest.raises(SerialCommunicationIOError):
        heinzinger.voltage = 1
    with pytest.raises(SerialCommunicationIOError):
        assert heinzinger.voltage
    with pytest.raises(SerialCommunicationIOError):
        heinzinger.current = 1
    with pytest.raises(SerialCommunicationIOError):
        assert heinzinger.current


def test_heinzinger_deprecation_warning(started_heinzinger_device) -> None:
    com, heinzinger = started_heinzinger_device
    with pytest.raises(DeprecationWarning):
        heinzinger.output_on()
    with pytest.raises(DeprecationWarning):
        heinzinger.output_off()
    with pytest.raises(DeprecationWarning):
        heinzinger.get_number_of_recordings()
    with pytest.raises(DeprecationWarning):
        heinzinger.set_number_of_recordings()
    with pytest.raises(DeprecationWarning):
        heinzinger.measure_voltage()
    with pytest.raises(DeprecationWarning):
        heinzinger.measure_current()
    com.put_text("0.0")  # when device stop, voltage is set to 0


# test different possible serial numbers to check if the info is read correctly
devices_data = [
    ("PNChp 60000-1neg 354211082", 60000, 0.001),
    ("314807440/PNChp 60000-1neg.", 60000, 0.001),
    ("PNChp 1500-40 ump. 375214277", 1500, 0.04),
    ("PNC 100000-6pos 375214277", 100000, 0.006),
    ("PNC 600-3000pos 375214277", 600, 3),
]


@pytest.mark.parametrize("device_data", devices_data)
def test_heinzinger_start(start_heinzinger_device, device_data) -> None:
    started_heinzinger = start_heinzinger_device(device_data[0])
    com, heinzinger = next(started_heinzinger)

    # starting the device again should work
    com.put_text("my_interface_version")
    com.put_text(device_data[0])
    heinzinger.start()
    assert com.get_written() == "VERS?"
    assert com.get_written() == "*IDN?"
    assert heinzinger.max_voltage_hardware == heinzinger.max_voltage == device_data[1]
    assert heinzinger.max_current_hardware == heinzinger.max_current == device_data[2]
    assert com.get_written() == "AVER " + str(
        int(heinzinger.config.default_number_of_recordings.value)
    )


def test_heinzinger_not_recognized(start_heinzinger_device) -> None:
    with pytest.raises(HeinzingerDeviceNotRecognizedError):
        next(start_heinzinger_device("ABC 600-3000pos 375214277"))


def test_heinzinger_set_max_voltage(start_heinzinger_device) -> None:
    started_heinzinger = start_heinzinger_device("PNC 600-3000pos 375214277")
    com, heinzinger = next(started_heinzinger)

    # set max_voltage to half the current value
    next_value = heinzinger.max_voltage / 2
    heinzinger.max_voltage = next_value
    assert heinzinger.max_voltage == next_value
    heinzinger.max_voltage = None

    # assigning value higher than hardware maximum
    with pytest.raises(ValueError):
        heinzinger.max_voltage = heinzinger.max_voltage_hardware * 2

    # assigning negative value
    with pytest.raises(ValueError):
        heinzinger.max_voltage = -1


def test_pnc_set_max_current(start_heinzinger_device) -> None:
    started_heinzinger = start_heinzinger_device("PNC 600-3000pos 375214277")
    com, heinzinger = next(started_heinzinger)

    # set max_voltage to half the current value
    next_value = heinzinger.max_current / 2
    heinzinger.max_current = next_value
    assert heinzinger.max_current == next_value
    heinzinger.max_current = None

    # assigning value higher than hardware maximum
    with pytest.raises(ValueError):
        heinzinger.max_current = heinzinger.max_current_hardware * 2

    # assigning negative value
    with pytest.raises(ValueError):
        heinzinger.max_current = -1


def test_pnc_set_negative(start_heinzinger_device) -> None:
    started_heinzinger = start_heinzinger_device("PNC 600-3000pos 375214277")
    com, heinzinger = next(started_heinzinger)

    # assigning negative value
    with pytest.raises(ValueError):
        heinzinger.current = -1
    with pytest.raises(ValueError):
        heinzinger.voltage = -1


# test different values for voltage and current
# (device_name, user input value, should results command, internal value, out-of-range
# value)
devices_u = [
    ("PNC 600-3000pos 375214277", 600, "VOLT 600.000000", 600, 700),
    ("PNChp 1500-40 ump. 375214277", 1, "VOLT 1.000000", 1, 1600),
    ("PNChp 60000-1neg 354211082", 50000, "VOLT 50000.000000", 50000, 60000.1),
    ("PNC 100000-6pos 375214277", 100000, "VOLT 100.000000", 100, 101000),
]


@pytest.mark.parametrize("device_data", devices_u)
def test_heinzinger_set_voltage(start_heinzinger_device, device_data) -> None:
    started_heinzinger = start_heinzinger_device(device_data[0])
    com, heinzinger = next(started_heinzinger)

    # test if the correct text is sent to the com
    com.put_text(str(device_data[3]))
    heinzinger.voltage = device_data[1]
    assert com.get_written().strip() == device_data[2]
    com.put_text(str(device_data[3]))
    assert heinzinger.voltage == device_data[1]
    com.put_text(str(device_data[3]))
    heinzinger.set_voltage = device_data[1]
    com.put_text(str(device_data[3]))
    assert heinzinger.set_voltage == device_data[1]

    # test if an error is raised when trying to set a too high voltage
    com.put_text(str(device_data[4]))
    with pytest.raises(ValueError):
        heinzinger.voltage = device_data[4]
    com.put_text("2.0")
    with pytest.raises(HeinzingerSetValueError):
        heinzinger.voltage = device_data[1]


# test different values for voltage and current
# (device_name, user input value, should results command, internal value, out-of-range
# value)
devices_i = [
    ("PNC 600-3000pos 375214277", 3, "CURR 3.000000", 3, 10),
    ("PNChp 1500-40 ump. 375214277", 0.039, "CURR 39.000000", 39, 0.041),
    ("PNChp 60000-1neg 354211082", 0.001, "CURR 1.000000", 1, 0.002),
    ("PNC 100000-6pos 375214277", 0, "CURR 0.000000", 0, 6.1),
]


@pytest.mark.parametrize("device_data", devices_i)
def test_heinzinger_set_current(start_heinzinger_device, device_data) -> None:
    started_heinzinger = start_heinzinger_device(device_data[0])
    com, heinzinger = next(started_heinzinger)

    # test if the correct text is sent to the com
    com.put_text(str(device_data[3]))
    heinzinger.current = device_data[1]
    assert com.get_written().strip() == device_data[2]
    com.put_text(str(device_data[3]))
    assert heinzinger.current == device_data[1]
    com.put_text(str(device_data[3]))
    heinzinger.set_current = device_data[1]
    com.put_text(str(device_data[3]))
    assert heinzinger.set_current == device_data[1]

    # test if an error is raised when trying to set a too high current
    com.put_text(str(device_data[4]))
    with pytest.raises(ValueError):
        heinzinger.current = device_data[4]
    com.put_text("2.0")
    with pytest.raises(HeinzingerSetValueError):
        heinzinger.current = device_data[1]



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Device class for controlling a Elektro Automatik PSI 9000 power supply over VISA.

It is necessary that a backend for pyvisa is installed.
This can be NI-Visa oder pyvisa-py (up to now, all the testing was done with NI-Visa)
"""

from .ea_psi9000 import (  # noqa: F401
    PSI9000,
    PSI9000Config,
    PSI9000Error,
    PSI9000VisaCommunication,
    PSI9000VisaCommunicationConfig,
)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Communication protocol for serial ports. Makes use of the `pySerial
<https://pythonhosted.org/pyserial/index.html>`_ library.
"""

import logging
import warnings
from typing import Optional, Union, cast

# Note: PyCharm does not recognize the dependency correctly, it is added as pyserial.
import serial

from hvl_ccb.comm import (
    AsyncCommunicationProtocol,
    AsyncCommunicationProtocolConfig,
    CommunicationError,
)
from hvl_ccb.configuration import ConfigurationValueWarning, configdataclass
from hvl_ccb.utils.enum import ValueEnum, unique
from hvl_ccb.utils.typing import Number

logger = logging.getLogger(__name__)


class SerialCommunicationIOError(IOError, CommunicationError):
    """Serial communication related I/O errors."""


@unique
class SerialCommunicationParity(ValueEnum):
    """
    Serial communication parity.
    """

    EVEN = serial.PARITY_EVEN
    MARK = serial.PARITY_MARK
    NAMES = serial.PARITY_NAMES
    NONE = serial.PARITY_NONE
    ODD = serial.PARITY_ODD
    SPACE = serial.PARITY_SPACE


@unique
class SerialCommunicationStopbits(ValueEnum):
    """
    Serial communication stopbits.
    """

    ONE = serial.STOPBITS_ONE
    ONE_POINT_FIVE = serial.STOPBITS_ONE_POINT_FIVE
    TWO = serial.STOPBITS_TWO


@unique
class SerialCommunicationBytesize(ValueEnum):
    """
    Serial communication bytesize.
    """

    FIVEBITS = serial.FIVEBITS
    SIXBITS = serial.SIXBITS
    SEVENBITS = serial.SEVENBITS
    EIGHTBITS = serial.EIGHTBITS


@configdataclass
class SerialCommunicationConfig(AsyncCommunicationProtocolConfig):
    """
    Configuration dataclass for :class:`SerialCommunication`.
    """

    Parity = SerialCommunicationParity
    Stopbits = SerialCommunicationStopbits
    Bytesize = SerialCommunicationBytesize

    #: Port is a string referring to a COM-port (e.g. ``'COM3'``) or a URL.
    #: The full list of capabilities is found `on the pyserial documentation
    #: <https://pythonhosted.org/pyserial/url_handlers.html>`_.
    port: Optional[str] = None

    #: Baudrate of the serial port
    baudrate: int = 9600

    #: Parity to be used for the connection.
    parity: Union[str, SerialCommunicationParity] = Parity.NONE

    #: Stopbits setting, can be 1, 1.5 or 2.
    stopbits: Union[Number, SerialCommunicationStopbits] = Stopbits.ONE

    #: Size of a byte, 5 to 8
    bytesize: Union[int, SerialCommunicationBytesize] = Bytesize.EIGHTBITS

    #: Timeout in seconds for the serial port
    timeout: Number = 2

    def clean_values(self) -> None:
        super().clean_values()

        if not isinstance(self.parity, SerialCommunicationParity):
            self.force_value("parity", SerialCommunicationParity(self.parity))  # type: ignore[attr-defined]

        if not isinstance(self.stopbits, SerialCommunicationStopbits):
            self.force_value("stopbits", SerialCommunicationStopbits(self.stopbits))  # type: ignore[attr-defined]

        if not isinstance(self.bytesize, SerialCommunicationBytesize):
            self.force_value("bytesize", SerialCommunicationBytesize(self.bytesize))  # type: ignore[attr-defined]

        if self.timeout < 0:
            msg = "Timeout has to be >= 0."
            raise ValueError(msg)
        if self.timeout < 1:  # min. viable threshold as tested w/ Arduino Nano board
            warnings.warn(
                "Setting a too low timeout for a serial connection communication"
                " may lead to random errors during a communication with a device.",
                category=ConfigurationValueWarning,
                stacklevel=2,
            )

    def create_serial_port(self) -> serial.Serial:
        """
        Create a serial port instance according to specification in this configuration

        :return: Closed serial port instance
        """

        ser = serial.serial_for_url(self.port, do_not_open=True)

        ser.baudrate = self.baudrate
        ser.parity = cast(SerialCommunicationParity, self.parity).value
        ser.stopbits = cast(SerialCommunicationStopbits, self.stopbits).value
        ser.bytesize = cast(SerialCommunicationBytesize, self.bytesize).value
        ser.timeout = self.timeout

        return ser

    def terminator_str(self) -> str:
        return self.terminator.decode()


class SerialCommunication(AsyncCommunicationProtocol):
    """
    Implements the Communication Protocol for serial ports.
    """

    def __init__(self, configuration) -> None:
        """
        Constructor for SerialCommunication.
        """

        super().__init__(configuration)

        self._serial_port = self.config.create_serial_port()

    @staticmethod
    def config_cls():
        return SerialCommunicationConfig

    def open(self) -> None:
        """
        Open the serial connection.

        :raises SerialCommunicationIOError: when communication port cannot be opened.
        """

        # open the port
        with self.access_lock:
            try:
                self._serial_port.open()
            except serial.SerialException as exc:
                # ignore when port is already open
                logger.exception(
                    "Error of Serial Connection, maybe it is already open.",
                    exc_info=exc,
                )
                if str(exc) != "Port is already open.":
                    raise SerialCommunicationIOError from exc

    def close(self) -> None:
        """
        Close the serial connection.
        """

        # close the port
        with self.access_lock:
            self._serial_port.close()

    @property
    def is_open(self) -> bool:
        """
        Flag indicating if the serial port is open.

        :return: `True` if the serial port is open, otherwise `False`
        """
        return self._serial_port.is_open

    def read_bytes(self) -> bytes:
        """
        Read the bytes from the serial port till the terminator is found.
        The input buffer may hold additional lines afterwards.

        This method uses `self.access_lock` to ensure thread-safety.

        :return: Bytes read from the serial port; `b''` if there was nothing to read.
        :raises SerialCommunicationIOError: when communication port is not opened
        """

        with self.access_lock:
            try:
                return self._serial_port.read_until(self.config.terminator)
            except serial.SerialException as exc:
                logger.exception("Error of Serial Communication", exc_info=exc)
                raise SerialCommunicationIOError from exc

    def write_bytes(self, data: bytes) -> int:
        """
        Write bytes to the serial port.

        This method uses `self.access_lock` to ensure thread-safety.

        :param data: data to write to the serial port
        :return: number of bytes written
        :raises SerialCommunicationIOError: when communication port is not opened
        """

        with self.access_lock:
            try:
                return self._serial_port.write(data)
            except serial.SerialException as exc:
                logger.exception("Error of Serial Communication", exc_info=exc)
                raise SerialCommunicationIOError from exc

    def read_single_bytes(self, size: int = 1) -> bytes:
        """
        Read the specified number of bytes from the serial port.
        The input buffer may hold additional data afterwards.

        :return: Bytes read from the serial port; `b''` if there was nothing to read.
        """

        with self.access_lock:
            try:
                return self._serial_port.read(size)
            except serial.SerialException as exc:
                logger.exception("Error of Serial Communication", exc_info=exc)
                raise SerialCommunicationIOError from exc



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
""" """

import logging
import time
from functools import wraps
from typing import Any, Callable, Optional, TypeVar, Union, cast

import libtiepie as ltp

# from libtiepie import i2chost as ltp_i2c
from libtiepie import generator as ltp_gen
from libtiepie import oscilloscope as ltp_osc
from libtiepie.exceptions import InvalidDeviceSerialNumberError, LibTiePieException

from hvl_ccb.configuration import configdataclass
from hvl_ccb.dev import DeviceError
from hvl_ccb.utils.enum import NameEnum
from hvl_ccb.utils.typing import Number

logger = logging.getLogger(__name__)


@configdataclass
class TiePieDeviceConfig:
    """
    Configuration dataclass for TiePie
    """

    serial_number: int
    require_block_measurement_support: bool = True
    n_max_try_get_device: int = 10
    wait_sec_retry_get_device: Number = 1.0
    is_data_ready_polling_interval_sec: Number = 0.01

    def clean_values(self) -> None:
        if self.serial_number <= 0:
            msg = "serial_number must be a positive integer."
            raise ValueError(msg)
        if self.n_max_try_get_device <= 0:
            msg = "n_max_try_get_device must be an positive integer."
            raise ValueError(msg)
        if self.wait_sec_retry_get_device <= 0:
            msg = "wait_sec_retry_get_device must be a positive number."
            raise ValueError(msg)
        if self.is_data_ready_polling_interval_sec <= 0:
            msg = "is_data_ready_polling_interval_sec must be a positive number."
            raise ValueError(msg)


class TiePieDeviceType(NameEnum, init="value ltp_class"):  # type: ignore[call-arg]
    """
    TiePie device type.
    """

    OSCILLOSCOPE = ltp.DEVICETYPE_OSCILLOSCOPE, ltp_osc.Oscilloscope
    # I2C = ltp.DEVICETYPE_I2CHOST, ltp_i2c.I2CHost
    GENERATOR = ltp.DEVICETYPE_GENERATOR, ltp_gen.Generator


class TiePieError(DeviceError):
    """
    Error of the class TiePie
    """


def wrap_libtiepie_exception(func: Callable) -> Callable:
    """
    Decorator wrapper for `libtiepie` methods that use
    `libtiepie.library.check_last_status_raise_on_error()` calls.

    :param func: Function or method to be wrapped
    :raises TiePieError: instead of `LibTiePieException` or one of its subtypes.
    :return: whatever `func` returns
    """

    @wraps(func)
    def wrapped_func(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except LibTiePieException as e:
            logger.exception("Error from LibTiePie")
            raise TiePieError from e

    return wrapped_func


_LtpDeviceReturnType = TypeVar("_LtpDeviceReturnType")
"""
An auxiliary typing hint of a `libtiepie` device type for return value of
the `get_device_by_serial_number` function and the wrapper methods using it.
"""


@wrap_libtiepie_exception
def get_device_by_serial_number(
    serial_number: int,
    # Note: TiePieDeviceType aenum as a tuple to define a return value type
    device_type: Union[str, tuple[int, _LtpDeviceReturnType]],
    n_max_try_get_device: int = 10,
    wait_sec_retry_get_device: float = 1.0,
) -> _LtpDeviceReturnType:
    """
    Open and return handle of TiePie device with a given serial number

    :param serial_number: int serial number of the device
    :param device_type: a `TiePieDeviceType` instance containing device identifier (int
        number) and its corresponding class, both from `libtiepie`, or a string name
        of such instance
    :param n_max_try_get_device: maximal number of device list updates (int number)
    :param wait_sec_retry_get_device: waiting time in seconds between retries (int
        number)
    :return: Instance of a `libtiepie`  device class according to the specified
        `device_type`
    :raises TiePieError: when there is no device with given serial number
    :raises ValueError: when `device_type` is not an instance of `TiePieDeviceType`
    """

    device_type = TiePieDeviceType(device_type)

    # include network search with ltp.device_list.update()
    ltp.network.auto_detect_enabled = True

    n_try = 0
    device_list_item: Optional[ltp.devicelistitem.DeviceListItem] = None
    while device_list_item is None and n_try < n_max_try_get_device:
        n_try += 1
        ltp.device_list.update()
        if not ltp.device_list:
            msg = f"Searching for device... (attempt #{n_try}/{n_max_try_get_device})"
            if n_try < n_max_try_get_device:
                logger.warning(msg)
                time.sleep(wait_sec_retry_get_device)
                continue
            msg = f"No devices found to start (attempt #{n_try}/{n_max_try_get_device})"
            logger.error(msg)
            raise TiePieError(msg)

        # if a device is found
        try:
            device_list_item = ltp.device_list.get_item_by_serial_number(serial_number)
        except InvalidDeviceSerialNumberError as e:
            msg = (
                f"The device with serial number {serial_number} is not "
                f"available; attempt #{n_try}/{n_max_try_get_device}."
            )
            if n_try < n_max_try_get_device:
                logger.warning(msg)
                time.sleep(wait_sec_retry_get_device)
                continue
            logger.exception(msg)
            raise TiePieError from e
    # assert device_list_item is not None  --> it cannot be None anymore

    if not cast(ltp.devicelistitem.DeviceListItem, device_list_item).can_open(
        device_type.value
    ):
        msg = (
            f"The device with serial number {serial_number} has no "
            f"{device_type} available."
        )
        logger.error(msg)
        raise TiePieError(msg)

    return cast(ltp.devicelistitem.DeviceListItem, device_list_item).open_device(
        device_type.value
    )


def _verify_via_libtiepie(
    dev_obj: ltp.device.Device, verify_method_suffix: str, value: Number
) -> Number:
    """
    Generic wrapper for `verify_SOMETHING` methods of the `libtiepie` device.
    Additionally to returning a value that will be actually set,
    gives an warning.

    :param dev_obj: TiePie device object, which has the verify_SOMETHING method
    :param verify_method_suffix: `libtiepie` devices verify_SOMETHING method
    :param value: numeric value
    :returns: Value that will be actually set instead of `value`.
    :raises TiePieError: when status of underlying device gives an error
    """
    verify_method = getattr(
        dev_obj,
        f"verify_{verify_method_suffix}",
    )
    will_have_value = verify_method(value)
    if will_have_value != value:
        if verify_method_suffix == "record_length":
            value_str = f"{value:_d}"
            set_value = f"{will_have_value:_d}"
        else:
            value_str = f"{value:_.3f}"
            set_value = f"{will_have_value:_.3f}"
        msg = (
            f"Can't set {verify_method_suffix} to "
            f"{value_str}; instead {set_value} will be set."
        )
        logger.warning(msg)
    return will_have_value


def _require_dev_handle(device_type) -> Callable[[Callable], Callable]:
    """
    Create method decorator to check if the TiePie device handle is available.

    :param device_type: the TiePie device type which device handle is required
    :raises ValueError: when `device_type` is not an instance of `TiePieDeviceType`
    """

    device_type: TiePieDeviceType = TiePieDeviceType(  # type: ignore[no-redef]
        device_type
    )

    def wrapper(method) -> Callable:
        """
        Method decorator to check if a TiePie device handle is available; raises
        `TiePieError` if hand is not available.

        :param method: `TiePieDevice` instance method to wrap
        :return: Whatever wrapped `method` returns
        """

        @wraps(method)
        def wrapped_func(self, *args, **kwargs) -> Any:
            dev_str = None
            if device_type is TiePieDeviceType.OSCILLOSCOPE and self._osc is None:
                dev_str = "oscilloscope"
            if device_type is TiePieDeviceType.GENERATOR and self._gen is None:
                dev_str = "generator"
            # if device_type is TiePieDeviceType.I2C and self._i2c is None:
            #     dev_str = "I2C host"
            if dev_str is not None:
                msg = f"The {dev_str} handle is not available; call `.start()` first."
                logger.error(msg)
                raise TiePieError(msg)
            return method(self, *args, **kwargs)

        return wrapped_func

    return wrapper



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Tests for TiePie
"""

import sys

import libtiepie as ltp
import numpy as np
import pytest
from libtiepie.exceptions import InvalidDeviceSerialNumberError, LibTiePieException
from pytest_mock import MockerFixture

from hvl_ccb.dev.tiepie import (  # TiePieI2CHostMixin,
    TiePieDeviceConfig,
    TiePieDeviceType,
    TiePieError,
    TiePieGeneratorMixin,
    TiePieGeneratorSignalType,
    TiePieHS5,
    TiePieHS6,
    TiePieOscilloscope,
    TiePieOscilloscopeAutoResolutionModes,
    TiePieOscilloscopeChannelCoupling,
    TiePieOscilloscopeRange,
    TiePieOscilloscopeResolution,
    TiePieOscilloscopeTriggerKind,
    TiePieOscilloscopeTriggerLevelMode,
    TiePieWS5,
    get_device_by_serial_number,
)
from hvl_ccb.dev.tiepie.base import _verify_via_libtiepie

# from mock_libtiepie import i2chost as _i2chost
from mock_libtiepie import generator as _generator
from mock_libtiepie import oscilloscope as _oscilloscope
from mock_libtiepie.const import (  # MOCK_I2CHOST_SERIAL_NUMBER,
    MOCK_DEVICE_SERIAL_NUMBER,
    MOCK_GENERATOR_SERIAL_NUMBER,
    MOCK_OSCILLOSCOPE_SERIAL_NUMBER,
    MOCK_OSCILLOSCOPE_SERIAL_NUMBER_2,
)
from mock_libtiepie.devicelist import device_list as _device_list

libtiepie = sys.modules["libtiepie"]

libtiepie.device_list = _device_list
libtiepie.oscilloscope = _oscilloscope
libtiepie.generator = _generator
# libtiepie.i2chost = _i2chost

sys.modules["libtiepie"] = libtiepie


@pytest.fixture(scope="module")
def com_config():
    return {}


@pytest.fixture(scope="module")
def dev_config():
    return {
        "serial_number": MOCK_DEVICE_SERIAL_NUMBER,
        "n_max_try_get_device": 2,
        "wait_sec_retry_get_device": 0.01,
    }


def test_instantiation(com_config, dev_config) -> None:
    dev_hs6 = TiePieHS6(com_config, dev_config)
    assert dev_hs6 is not None
    dev_hs5 = TiePieHS5(com_config, dev_config)
    assert dev_hs5 is not None
    dev_ws5 = TiePieWS5(com_config, dev_config)
    assert dev_ws5 is not None


def test_wrap_libtiepie_exception(
    com_config, dev_config, mocker: MockerFixture
) -> None:
    dev_osc = TiePieOscilloscope(com_config, dev_config)

    def raise_libtiepie_exception():
        raise LibTiePieException(0, "mock")

    mocker.patch(
        "libtiepie.device_list.update",
        side_effect=raise_libtiepie_exception,
        autospec=True,
    )

    with pytest.raises(TiePieError):
        dev_osc.list_devices()

    with pytest.raises(TiePieError):
        get_device_by_serial_number(
            dev_osc.config.serial_number, TiePieDeviceType.OSCILLOSCOPE
        )

    with pytest.raises(TiePieError):
        dev_osc.start()


def test_list_devices(com_config, dev_config) -> None:
    dev_osc = TiePieOscilloscope(com_config, dev_config)
    list_devices = dev_osc.list_devices()
    assert list_devices is not None


def test_list_devices_none_available(
    com_config, dev_config, mocker: MockerFixture
) -> None:
    # simulate absence of devices
    mock_devices = mocker.patch.object(libtiepie.device_list, "mock_devices")
    mock_devices.return_value = False

    dev_osc = TiePieOscilloscope(com_config, dev_config)
    list_devices = dev_osc.list_devices()
    assert not list_devices


def test_get_device_by_serial_number() -> None:
    dev_osc = get_device_by_serial_number(
        MOCK_OSCILLOSCOPE_SERIAL_NUMBER,
        TiePieDeviceType.OSCILLOSCOPE,
    )
    assert dev_osc is not None

    dev_gen = get_device_by_serial_number(
        MOCK_GENERATOR_SERIAL_NUMBER,
        TiePieDeviceType.GENERATOR,
    )
    assert dev_gen is not None

    # dev_i2c = get_device_by_serial_number(
    #     MOCK_I2CHOST_SERIAL_NUMBER,
    #     TiePieDeviceType.I2C,
    # )
    # assert dev_i2c is not None


@pytest.mark.parametrize(
    "wrong_dev_config",
    [
        {"serial_number": -23},
        {"n_max_try_get_device": 0},
        {"wait_sec_retry_get_device": 0.0},
        {"is_data_ready_polling_interval_sec": 0.0},
    ],
)
def test_invalid_config_dict(dev_config, wrong_dev_config) -> None:
    invalid_config = dict(dev_config)
    invalid_config.update(wrong_dev_config)
    with pytest.raises(ValueError):
        TiePieDeviceConfig(**invalid_config)


def test_get_device_by_serial_number_not_available(
    dev_config, mocker: MockerFixture
) -> None:
    # simulate absence of devices
    mock_devices = mocker.patch.object(libtiepie.device_list, "mock_devices")
    mock_devices.return_value = False

    with pytest.raises(TiePieError):
        get_device_by_serial_number(
            MOCK_OSCILLOSCOPE_SERIAL_NUMBER,
            TiePieDeviceType.OSCILLOSCOPE,
            n_max_try_get_device=dev_config["n_max_try_get_device"],
            wait_sec_retry_get_device=dev_config["wait_sec_retry_get_device"],
        )


def test_get_device_by_serial_number_wrong_device_type() -> None:
    with pytest.raises(TiePieError):
        get_device_by_serial_number(
            MOCK_GENERATOR_SERIAL_NUMBER,
            TiePieDeviceType.OSCILLOSCOPE,
        )


def test_get_device_by_serial_number_invalid_device_sn(
    dev_config,
    mocker: MockerFixture,
) -> None:
    # simulate invalid device serian number error
    get_item_by_serial_number = mocker.patch.object(
        libtiepie.device_list,
        "get_item_by_serial_number",
    )
    get_item_by_serial_number.side_effect = InvalidDeviceSerialNumberError

    with pytest.raises(TiePieError):
        get_device_by_serial_number(
            MOCK_OSCILLOSCOPE_SERIAL_NUMBER,
            TiePieDeviceType.OSCILLOSCOPE,
            n_max_try_get_device=dev_config["n_max_try_get_device"],
            wait_sec_retry_get_device=dev_config["wait_sec_retry_get_device"],
        )


def test_block_measurement_support() -> None:
    dev_osc = TiePieOscilloscope(
        {},
        {"serial_number": MOCK_OSCILLOSCOPE_SERIAL_NUMBER_2},
    )

    with pytest.raises(TiePieError):
        dev_osc.start()


def test_channel_config_on_start_stop(com_config, dev_config) -> None:
    dev_osc = TiePieOscilloscope(com_config, dev_config)

    assert not dev_osc.config_osc_channel_dict
    # with pytest.raises(TiePieError):
    #     dev_osc.n_channels

    dev_osc.start()

    n_channels = dev_osc.n_channels

    for ch_nr in range(1, n_channels + 1):
        assert ch_nr in dev_osc.config_osc_channel_dict
        assert not dev_osc.config_osc_channel_dict[ch_nr].enabled

    for wrong_ch_nr in (0, dev_osc.n_channels + 1):
        with pytest.raises(KeyError):
            dev_osc.config_osc_channel_dict[wrong_ch_nr]

    dev_osc.stop()


def test_channels_enabled(com_config, dev_config) -> None:
    dev_osc = TiePieOscilloscope(com_config, dev_config)
    dev_osc.start()
    for ch_config in dev_osc.config_osc_channel_dict.values():
        ch_config.enabled = True
    assert list(dev_osc.channels_enabled) == list(
        dev_osc.config_osc_channel_dict.keys()
    )
    dev_osc.config_osc_channel_dict[1].enabled = False
    dev_osc.stop()


def test_n_channels(com_config, dev_config) -> None:
    dev_osc = TiePieOscilloscope(com_config, dev_config)
    dev_osc.start()
    assert dev_osc.n_channels == 3
    dev_osc.stop()


def test_dev_osc_not_running(com_config, dev_config) -> None:
    dev_osc = TiePieOscilloscope(com_config, dev_config)
    with pytest.raises(TiePieError):
        assert dev_osc.n_channels == 3

    class TiePieGeneratorOscilloscope(TiePieGeneratorMixin, TiePieOscilloscope):
        pass

    dev_gen = TiePieGeneratorOscilloscope(com_config, dev_config)
    with pytest.raises(TiePieError):
        dev_gen.generator_start()

    # class TiePieI2CHostOscilloscope(TiePieI2CHostMixin, TiePieOscilloscope):
    #     pass
    #
    # dev_i2c = TiePieI2CHostOscilloscope(com_config, dev_config)
    #
    # with pytest.raises(TiePieError):
    #     dev_i2c._i2c_config_setup()  # ATM only priv method avail for testing


def test_safeground_enabled(com_config, dev_config) -> None:
    dev_osc = TiePieOscilloscope(com_config, dev_config)

    dev_osc.start()

    for ch_config in dev_osc.config_osc_channel_dict.values():
        assert ch_config.safeground_enabled is False

    for ch_nr in [1, 3]:
        dev_osc.config_osc_channel_dict[ch_nr].safeground_enabled = True
    for ch_nr in [2]:
        assert dev_osc.config_osc_channel_dict[ch_nr].safeground_enabled is False

    dev_osc.stop()


def test_safeground_not_available(com_config, dev_config) -> None:
    dev_osc = TiePieOscilloscope(com_config, dev_config)

    dev_osc.start()

    for ch_config in dev_osc.config_osc_channel_dict.values():
        ch_config._channel.disable_safeground_option()

    for ch_nr in dev_osc.config_osc_channel_dict:
        with pytest.raises(TiePieError):
            dev_osc.config_osc_channel_dict[ch_nr].safeground_enabled = True
        with pytest.raises(TiePieError):
            assert dev_osc.config_osc_channel_dict[ch_nr].safeground_enabled

    dev_osc.stop()


def test_trigger_enabled(com_config, dev_config) -> None:
    dev_osc = TiePieOscilloscope(com_config, dev_config)

    for ch_config in dev_osc.config_osc_channel_dict.values():
        ch_config.trigger_enabled = False
        assert ch_config.trigger_enabled is False

    dev_osc.start()

    for ch_config in dev_osc.config_osc_channel_dict.values():
        assert ch_config.trigger_enabled is False

    for ch_nr in [1, 3]:
        dev_osc.config_osc_channel_dict[ch_nr].trigger_enabled = True
        assert dev_osc.config_osc_channel_dict[ch_nr].trigger_enabled is True

    for ch_nr in [2]:
        assert dev_osc.config_osc_channel_dict[ch_nr].trigger_enabled is False

    for ch_nr in [3]:
        dev_osc.config_osc_channel_dict[ch_nr].trigger_enabled = False

    for ch_nr in [1]:
        assert dev_osc.config_osc_channel_dict[ch_nr].trigger_enabled is True

    for ch_nr in [2, 3]:
        assert dev_osc.config_osc_channel_dict[ch_nr].trigger_enabled is False

    dev_osc.stop()


def test_scope_config_set(com_config, dev_config) -> None:
    dev_osc = TiePieOscilloscope(com_config, dev_config)

    assert dev_osc.config_osc is None

    dev_osc.start()

    config_osc = dev_osc.config_osc

    # record length
    record_length_value = 100
    config_osc.record_length = record_length_value
    assert config_osc.record_length == record_length_value

    record_length_value = 1e6
    config_osc.record_length = record_length_value
    assert config_osc.record_length == record_length_value

    # sample rate
    sample_rate_value = 1e6
    config_osc.sample_rate = sample_rate_value
    assert config_osc.sample_rate == sample_rate_value

    # pre-sample ratio
    pre_sample_ratio_value = 0.2
    config_osc.pre_sample_ratio = pre_sample_ratio_value
    assert config_osc.pre_sample_ratio == pre_sample_ratio_value

    # resolution
    resolution = TiePieOscilloscopeResolution.FOURTEEN_BIT
    config_osc.resolution = resolution
    assert config_osc.resolution == resolution

    # trigger timeout
    trigger_timeouts = [0, 1, 1.5]
    for trigger_timeout in trigger_timeouts:
        config_osc.trigger_timeout = trigger_timeout
        assert config_osc.trigger_timeout == trigger_timeout

    trigger_timeouts = [None, -1, ltp.TO_INFINITY]
    for trigger_timeout in trigger_timeouts:
        config_osc.trigger_timeout = trigger_timeout
        assert config_osc.trigger_timeout is None

    # auto resolution mode
    auto_resolution_mode = TiePieOscilloscopeAutoResolutionModes.DISABLED
    config_osc.auto_resolution_mode = auto_resolution_mode
    assert config_osc.auto_resolution_mode == auto_resolution_mode

    dev_osc.stop()

    assert dev_osc.config_osc is None


def test_scope_config_set_invalid(com_config, dev_config) -> None:
    dev_osc = TiePieOscilloscope(com_config, dev_config)

    dev_osc.start()

    config_osc = dev_osc.config_osc

    # sample_rate
    with pytest.raises(ValueError):
        config_osc.sample_rate = -1
    with pytest.raises(ValueError):
        config_osc.sample_rate = 1e14
    with pytest.raises(TypeError):
        config_osc.sample_rate = "1"

    # record_length
    with pytest.raises(ValueError):
        config_osc.record_length = -1
    with pytest.raises(TypeError):
        config_osc.record_length = "1"
    with pytest.raises(ValueError):
        config_osc.record_length = 1.5

    # pre_sample_ratio
    for v in (-0.1, 1.1):
        with pytest.raises(ValueError):
            config_osc.pre_sample_ratio = v
    with pytest.raises(TypeError):
        config_osc.pre_sample_ratio = "0.5"

    # resolution
    with pytest.raises(ValueError):
        config_osc.resolution = 1
    with pytest.raises(TypeError):
        config_osc.resolution = "8"

    # trigger_timeout
    with pytest.raises(ValueError):
        config_osc.trigger_timeout = -0.1
    with pytest.raises(ValueError):
        config_osc.trigger_timeout = -1.1
    with pytest.raises(TypeError):
        config_osc.trigger_timeout = "123"

    # auto resolution mode
    with pytest.raises(TypeError):
        config_osc.auto_resolution_mode = False
    with pytest.raises(TypeError):
        config_osc.auto_resolution_mode = "Disabled"


def test_channel_config_set(com_config, dev_config) -> None:
    dev_osc = TiePieOscilloscope(com_config, dev_config)

    assert not dev_osc.config_osc_channel_dict

    dev_osc.start()

    assert dev_osc.config_osc_channel_dict

    # Channel 2
    ch2_config = dev_osc.config_osc_channel_dict[2]

    # coupling
    coupling_value = TiePieOscilloscopeChannelCoupling.ACV
    ch2_config.coupling = coupling_value
    assert ch2_config.coupling == coupling_value

    ch2_config.coupling = "ACV"
    assert ch2_config.coupling is TiePieOscilloscopeChannelCoupling.ACV

    # enabled
    enabled_value = True
    ch2_config.enabled = enabled_value
    assert ch2_config.enabled == enabled_value

    # range
    range_value = TiePieOscilloscopeRange.TWENTY_VOLT
    ch2_config.input_range = range_value
    assert ch2_config.input_range == range_value

    for v, ir in (
        (20, TiePieOscilloscopeRange.TWENTY_VOLT),
        (21, TiePieOscilloscopeRange.FORTY_VOLT),
        (79.9, TiePieOscilloscopeRange.EIGHTY_VOLT),
    ):
        ch2_config.input_range = v
        assert ch2_config.input_range is ir

    # probe offset
    with pytest.raises(NotImplementedError):
        ch2_config.probe_offset = 1
    with pytest.raises(NotImplementedError):
        assert ch2_config.probe_offset == 1

    # probe gain
    with pytest.raises(NotImplementedError):
        ch2_config.probe_gain = 1
    with pytest.raises(NotImplementedError):
        assert ch2_config.probe_gain == 1

    # safeground enabled
    ch2_config._channel.enable_safeground_option()
    safeground_enabled_value = True
    ch2_config.safeground_enabled = safeground_enabled_value
    assert ch2_config.safeground_enabled == safeground_enabled_value
    ch2_config.safeground_enabled = False
    assert ch2_config.safeground_enabled is False

    # trigger level mode
    trigger_level_mode_value = TiePieOscilloscopeTriggerLevelMode.ABSOLUTE
    ch2_config.trigger_level_mode = trigger_level_mode_value
    assert ch2_config.trigger_level_mode == trigger_level_mode_value

    # trigger level
    trigger_level_value = 0.5
    ch2_config.trigger_level = trigger_level_value
    assert ch2_config.trigger_level == trigger_level_value

    trigger_level_value = 1
    ch2_config.trigger_level = trigger_level_value
    assert ch2_config.trigger_level == trigger_level_value

    # trigger level mode
    trigger_level_mode_value = TiePieOscilloscopeTriggerLevelMode.RELATIVE
    ch2_config.trigger_level_mode = trigger_level_mode_value
    assert ch2_config.trigger_level_mode == trigger_level_mode_value

    # trigger level
    trigger_level_value = 0.5
    ch2_config.trigger_level = trigger_level_value
    assert ch2_config.trigger_level == trigger_level_value

    # trigger hysteresis
    trigger_hysteresis_value = 0.05
    ch2_config.trigger_hysteresis = trigger_hysteresis_value
    assert ch2_config.trigger_hysteresis == trigger_hysteresis_value

    # trigger kind
    trigger_kind_value = TiePieOscilloscopeTriggerKind.ANY
    ch2_config.trigger_kind = trigger_kind_value
    assert ch2_config.trigger_kind == trigger_kind_value

    ch2_config.trigger_kind = "FALLING"
    assert ch2_config.trigger_kind is TiePieOscilloscopeTriggerKind.FALLING

    # trigger enabled
    trigger_enabled_value = True
    ch2_config.trigger_enabled = trigger_enabled_value
    assert ch2_config.trigger_enabled == trigger_enabled_value

    dev_osc.stop()

    assert not dev_osc.config_osc_channel_dict


def test_channel_config_set_invalid(com_config, dev_config) -> None:
    dev_osc = TiePieOscilloscope(com_config, dev_config)

    assert not dev_osc.config_osc_channel_dict

    dev_osc.start()

    # Channel 2
    ch_config = dev_osc.config_osc_channel_dict[2]

    # coupling
    for v in ("dcv", 123):
        with pytest.raises(ValueError):
            ch_config.coupling = v

    # enabled
    for v in ("True", 1):
        with pytest.raises(TypeError):
            ch_config.enabled = v

    # input_range
    with pytest.raises(TypeError):
        ch_config.input_range = "40"

    with pytest.raises(ValueError):
        ch_config.input_range = 234567

    # safeground_enabled
    for v in ("True", 1):
        with pytest.raises(TypeError):
            ch_config.safeground_enabled = v

    # trigger_hysteresis
    for v in (-0.1, 1.1):
        with pytest.raises(ValueError):
            ch_config.trigger_hysteresis = v
    with pytest.raises(TypeError):
        ch_config.trigger_hysteresis = "0.5"

    # trigger_kind
    for v in ("any", 123):
        with pytest.raises(ValueError):
            ch_config.trigger_kind = v

    # trigger_enabled
    for v in ("True", 1):
        with pytest.raises(TypeError):
            ch_config.trigger_enabled = v


def test_trigger_check(com_config, dev_config) -> None:
    dev_osc = TiePieOscilloscope(com_config, dev_config)

    dev_osc.start()

    assert dev_osc.is_triggered() is False
    dev_osc.force_trigger()
    dev_osc.config_osc.dev_osc.is_triggered = True
    assert dev_osc.is_triggered() is True

    dev_osc.stop()


_record_length = 7
_valid_pre_sample_count = 3
_sample_rate = 1000
_pre_sample_ratio = 0.5


def _test_measure_data(data, rec_lngth_reduced=0):
    assert len(data) == _record_length - rec_lngth_reduced
    for i, ch_data in enumerate(data.T):
        _test_measure_data_ch(i + 1, ch_data, rec_lngth_reduced=rec_lngth_reduced)


def _test_measure_data_ch(ch_nr, ch_data, rec_lngth_reduced=0):
    assert len(ch_data) == _record_length - rec_lngth_reduced
    for i in range(_record_length - rec_lngth_reduced):
        if ch_nr == 1:  # Check time vector
            assert ch_data[i] == (i - _valid_pre_sample_count) / _sample_rate
        else:
            assert ch_data[i] == float(i)


def test_start_and_stop_measurement(com_config, dev_config) -> None:
    dev_osc = TiePieOscilloscope(com_config, dev_config)
    dev_osc.start()
    if not dev_osc.is_measurement_running():
        dev_osc.start_measurement()
    dev_osc._osc.is_running = True
    assert dev_osc.is_measurement_running()
    with pytest.raises(TiePieError):
        dev_osc.start_measurement()
    dev_osc.stop_measurement()
    dev_osc._osc.is_running = False
    assert not dev_osc.is_measurement_running()
    with pytest.raises(TiePieError):
        dev_osc.stop_measurement()
    dev_osc.stop()


def _test_measure_and_collect(dev_osc, timeout, rec_lngth_reduced=0):
    # Enable all channels for measurement
    for ch in dev_osc.config_osc_channel_dict.values():
        ch.enabled = True

    dev_osc.config_osc.dev_osc.record_length = _record_length + 1
    dev_osc.config_osc.dev_osc.valid_pre_sample_count = _valid_pre_sample_count
    dev_osc.config_osc.dev_osc.sample_rate = _sample_rate
    dev_osc.config_osc.dev_osc.pre_sample_ratio = _pre_sample_ratio

    # Let the oscilloscope gather data
    dev_osc.start_measurement()

    if dev_osc.is_measurement_data_ready():
        data = dev_osc.collect_measurement_data(timeout)

    assert all(isinstance(x, np.ndarray) for x in data)
    _test_measure_data(data, rec_lngth_reduced=rec_lngth_reduced)


def test_measure_and_collect(com_config, dev_config) -> None:
    dev_osc = TiePieOscilloscope(com_config, dev_config)

    dev_osc.start()

    timeout = 0  # Same as default option
    _test_measure_and_collect(dev_osc, timeout)

    timeout = 0.2
    _test_measure_and_collect(dev_osc, timeout)

    timeout = None
    _test_measure_and_collect(dev_osc, timeout)

    timeout = 0  # Same as default option, but truncated data
    dev_osc.config_osc.dev_osc.mock_reduce_data = 1
    _test_measure_and_collect(dev_osc, timeout, rec_lngth_reduced=1)

    dev_osc.stop()


def test_measure_and_collect_fail(com_config, dev_config) -> None:
    dev_osc = TiePieOscilloscope(com_config, dev_config)

    dev_osc.start()

    # string as timeout
    timeout = "a"
    with pytest.raises(ValueError):
        dev_osc.collect_measurement_data(timeout)

    timeout = 0.2
    dev_osc.config_osc.dev_osc.is_data_ready = False
    assert dev_osc.collect_measurement_data(timeout) is None

    dev_osc.config_osc = None
    assert dev_osc.collect_measurement_data(timeout) is None
    assert dev_osc._check_record_length(np.zeros(1)) is None

    dev_osc.stop()


def test_generator_config_set(com_config, dev_config) -> None:
    dev_ws5 = TiePieWS5(com_config, dev_config)
    assert dev_ws5 is not None

    assert dev_ws5.config_gen is None

    dev_ws5.start()

    config_gen = dev_ws5.config_gen

    # frequency
    frequency_value = 1e6
    config_gen.frequency = frequency_value
    assert config_gen.frequency == frequency_value

    # amplitude
    amplitude_value = 2
    config_gen.amplitude = amplitude_value
    assert config_gen.amplitude == amplitude_value

    # offset
    offset_value = 1
    config_gen.offset = offset_value
    assert config_gen.offset == offset_value

    # signal type
    signal_type = TiePieGeneratorSignalType.SINE
    config_gen.signal_type = signal_type
    assert config_gen.signal_type == signal_type

    # arbitrary waveform
    t_axis = np.linspace(0, 100, 8192)  # Create signal array
    waveform = np.sin(t_axis) * (1 - t_axis / 100)
    config_gen.waveform = waveform
    assert config_gen.waveform.all() == waveform.all()

    # Create 2dim waveform
    waveform = np.array([waveform, waveform])
    with pytest.raises(ValueError):
        config_gen.waveform = waveform

    # Create signal array that is too large
    t_axis = np.linspace(0, 100, 10_001)
    waveform = np.sin(t_axis) * (1 - t_axis / 100)
    with pytest.raises(ValueError):
        config_gen.waveform = waveform

    # enabled
    enabled_value = True
    config_gen.enabled = enabled_value
    assert config_gen.enabled == enabled_value
    config_gen.enabled = False
    assert config_gen.enabled is False

    dev_ws5.stop()

    assert dev_ws5.config_gen is None


def test_generator_config_set_invalid(com_config, dev_config) -> None:
    dev_ws5 = TiePieWS5(com_config, dev_config)
    assert dev_ws5 is not None

    dev_ws5.start()

    config_gen = dev_ws5.config_gen

    # frequency
    with pytest.raises(ValueError):
        config_gen.frequency = -1
    with pytest.raises(ValueError):
        config_gen.frequency = 1e14
    with pytest.raises(TypeError):
        config_gen.frequency = "1"

    # amplitude
    with pytest.raises(ValueError):
        config_gen.amplitude = -1
    with pytest.raises(ValueError):
        config_gen.amplitude = 100
    with pytest.raises(TypeError):
        config_gen.frequency = "1"

    # offset
    with pytest.raises(ValueError):
        config_gen.offset = 100
    with pytest.raises(TypeError):
        config_gen.offset = "1"

    # signal type
    with pytest.raises(ValueError):
        config_gen.signal_type = "sin"

    # arbitrary waveform
    with pytest.raises(ValueError):
        config_gen.waveform = [1, 2, 3]

    # enabled
    for v in ("True", 1):
        with pytest.raises(TypeError):
            config_gen.enabled = v

    dev_ws5.stop()


def test_generate_signal(com_config, dev_config) -> None:
    dev_ws5 = TiePieWS5(com_config, dev_config)
    assert dev_ws5 is not None

    dev_ws5.start()

    assert not dev_ws5.generator_is_running

    dev_ws5.generator_start()
    dev_ws5.generator_start()
    assert dev_ws5.generator_is_running
    dev_ws5.generator_stop()
    dev_ws5.generator_stop()

    assert not dev_ws5.generator_is_running

    dev_ws5.stop()


def test_verify_via_libtiepie() -> None:
    dev_osc = get_device_by_serial_number(
        MOCK_OSCILLOSCOPE_SERIAL_NUMBER,
        TiePieDeviceType.OSCILLOSCOPE,
    )
    assert dev_osc is not None
    assert _verify_via_libtiepie(dev_osc, "sample_rate", 1e6) == 1e6
    assert _verify_via_libtiepie(dev_osc, "record_length", 10) == 10


def test_config_print_values(com_config, dev_config) -> None:
    dev_osc = TiePieOscilloscope(com_config, dev_config)
    dev_osc.start()
    config_osc = dev_osc.config_osc
    config_osc.record_length = 100
    assert f"record_length={config_osc.record_length!r}" in str(config_osc)

    ch_config = dev_osc.config_osc_channel_dict[1]
    assert f"enabled={ch_config.enabled!r}" in str(ch_config)
    ch_config.enabled = not ch_config.enabled
    assert f"enabled={ch_config.enabled!r}" in str(ch_config)

    dev_ws5 = TiePieWS5(com_config, dev_config)
    dev_ws5.start()
    config_gen = dev_ws5.config_gen
    config_gen.amplitude = 12
    assert f"amplitude={config_gen.amplitude!r}" in str(config_gen)

    # config_i2c = dev_ws5.config_i2c
    # assert "()" in str(config_i2c)  # no config properties yet


def test_missing_i2c_support() -> None:
    with pytest.raises(DeprecationWarning):
        from hvl_ccb.dev.tiepie import i2c  # noqa: F401



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Constants, ValueEnum: MeasurementFunction and TriggerSoruce
Descriptors for range, filter and aperture

TODO: Fix typing
"""

import logging

from hvl_ccb.dev import DeviceError
from hvl_ccb.dev.fluke884x.ranges import (
    ACCurrentRange,
    ACVoltageRange,
    ApertureRange,
    DCCurrentRange,
    DCVoltageRange,
    FilterRange,
    ResistanceRange,
)
from hvl_ccb.utils.enum import ValueEnum
from hvl_ccb.utils.validation import validate_number

logger = logging.getLogger(__name__)


class Fluke8845aError(DeviceError):
    pass


class Fluke8845aCheckError(Fluke8845aError):
    pass


class Fluke8845aUnknownCommandError(Fluke8845aError):
    pass


class MeasurementFunction(ValueEnum):
    """
    Page 40

    Sets the Meter function. This command must be followed by the INIT and FETCh?
    commands to cause the meter to take a measurement.
    """

    CURRENT_AC = "CURR:AC"
    CURRENT_DC = "CURR"
    VOLTAGE_AC = "VOLT:AC"
    VOLTAGE_DC = "VOLT"
    FOUR_WIRE_RESISTANCE = "FRES"
    TWO_WIRE_RESISTANCE = "RES"
    FREQUENCY = "FREQ"
    PERIOD = "PER"
    DIODE = "DIOD"

    def _range(self):
        if self == MeasurementFunction.VOLTAGE_AC:
            return ACVoltageRange
        if self == MeasurementFunction.VOLTAGE_DC:
            return DCVoltageRange
        if self == MeasurementFunction.CURRENT_AC:
            return ACCurrentRange
        if self == MeasurementFunction.CURRENT_DC:
            return DCCurrentRange
        if self in (
            MeasurementFunction.TWO_WIRE_RESISTANCE,
            MeasurementFunction.FOUR_WIRE_RESISTANCE,
        ):
            return ResistanceRange
        return None

    def _filter(self):
        if self in (MeasurementFunction.VOLTAGE_AC, MeasurementFunction.CURRENT_AC):
            return FilterRange
        return None

    def _aperture(self):
        if self in (MeasurementFunction.FREQUENCY, MeasurementFunction.PERIOD):
            return ApertureRange
        return None


class TriggerSource(ValueEnum):
    """
    Page 57

    .. code-block:: rst

        BUS: Sets the Meter to expect a trigger through the IEEE-488 bus or
        upon execution of a *TRG command
        IMM: Selects Meter's internal triggering system
        EXT: Sets the Meter to sense triggers through the trigger jack on the rear
        panel of the Meter
    """

    BUS = "BUS"
    IMMEDIATE = "IMM"
    EXTERNAL = "EXT"


class _BaseDescriptor:
    def __init__(self, cmd_enum):  # noqa: ANN204
        self._cmd_enum = cmd_enum
        self._query_cmd = None
        self._write_cmd = None
        self._range_enum = None
        self._value_min = None
        self._value_max = None

    def __get__(self, instance, owner=None):  # noqa: ANN204
        voltage_range = float(instance.com.query(self._query_cmd))
        return self._range_enum(voltage_range)

    def __set__(self, instance, input_range):  # noqa: ANN204
        if not isinstance(input_range, self._range_enum):
            validate_number(
                "input range",
                input_range,
                (self._value_min, self._value_max),
                logger=logger,
            )
        input_range = self._range_enum(input_range)
        instance.com.write(f"{self._write_cmd} {input_range}")
        if self.__get__(instance) == float(input_range.value):
            logger.info(
                f"{self._write_cmd} is successfully set "
                f"to {input_range} {self._range_enum.unit()}."
            )
        else:
            msg = f"{self._cmd_enum} setting failed."
            logger.error(msg)
            raise Fluke8845aCheckError(msg)


class _RangeDescriptor(_BaseDescriptor):
    def __init__(self, cmd_enum):  # noqa: ANN204
        super().__init__(cmd_enum)
        self._query_cmd = f"{self._cmd_enum}:RANG?"
        self._write_cmd = f"{self._cmd_enum}:RANG"
        self._range_enum = cmd_enum._range()
        self._value_min = 0
        self._value_max = max(self._range_enum)

    def __set__(self, instance, input_range):  # noqa: ANN204
        super().__set__(instance, input_range)


class _FilterDescriptor(_BaseDescriptor):
    def __init__(self, cmd_enum):  # noqa: ANN204
        super().__init__(cmd_enum)
        self._query_cmd = f"{self._cmd_enum}:BAND?"
        self._write_cmd = f"{self._cmd_enum}:BAND"
        self._range_enum = cmd_enum._filter()
        self._value_min = min(self._range_enum)
        self._value_max = max(self._range_enum)


class _ApertureDescriptor(_BaseDescriptor):
    def __init__(self, cmd_enum):  # noqa: ANN204
        super().__init__(cmd_enum)
        self._query_cmd = f"{self._cmd_enum}:APER?"
        self._write_cmd = f"{self._cmd_enum}:APER"
        self._range_enum = cmd_enum._aperture()
        self._value_min = min(self._range_enum)
        self._value_max = max(self._range_enum)



# #  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
# #
# """
# Mock I2C host
# """
# from libtiepie.i2chost import I2CHost as LtpI2CHost
#
#
# class I2CHost(LtpI2CHost):
#     """"""
#
#     def __init__(self, serial_number):
#         self._serial_number = serial_number



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Mock constants
"""

MOCK_DEVICE_SERIAL_NUMBER = 34937  # Device with oscilloscope, generator and i2c
MOCK_I2CHOST_SERIAL_NUMBER = 35492  # Device with I2C host only
MOCK_GENERATOR_SERIAL_NUMBER = 23456  # Device with generator only
MOCK_OSCILLOSCOPE_SERIAL_NUMBER = 12345  # Device with oscilloscope only
MOCK_OSCILLOSCOPE_SERIAL_NUMBER_2 = 42345  # Oscilloscope without block measurement

MMB_BLOCK = 1
MM_BLOCK = 1 << MMB_BLOCK
